from nipype.interfaces import fsl
from nipype.interfaces.spm.preprocess import Coregister
from banana.citation import spm_cite
from banana.file_format import (
    nifti_format, motion_mats_format, nifti_gz_format,
    multi_nifti_gz_format, zip_format, STD_IMAGE_FORMATS)
from arcana.data import FilesetSpec, FieldSpec, InputFilesetSpec
from banana.study import Study, StudyMetaClass
from banana.citation import fsl_cite, bet_cite, bet2_cite, ants_cite
from banana.file_format import (
    dicom_format, gif_format, nifti_gz_x_format)
from nipype.interfaces.utility import IdentityInterface
from banana.requirement import fsl_req, mrtrix_req, ants_req, spm_req, c3d_req
from nipype.interfaces.fsl import (FLIRT, FNIRT, Reorient2Std)
from arcana.exceptions import (
    ArcanaOutputNotProducedException, ArcanaMissingDataException)
from banana.interfaces.mrtrix.transform import MRResize
from banana.interfaces.custom.dicom import (
    DicomHeaderInfoExtraction, NiftixHeaderInfoExtraction)
from nipype.interfaces.utility import Merge
from nipype.interfaces import ants
from banana.interfaces.fsl import FSLSlices
from banana.file_format import text_matrix_format
import logging
from banana.interfaces.ants import AntsRegSyn
from banana.interfaces.custom.coils import ToPolarCoords
from arcana.utils.interfaces import ListDir, CopyToDir
from nipype.interfaces.ants.resampling import ApplyTransforms
from nipype.interfaces.fsl.preprocess import ApplyXFM
from banana.interfaces.c3d import ANTs2FSLMatrixConversion
from arcana import ParamSpec, SwitchSpec
from banana.interfaces.custom.motion_correction import (
    MotionMatCalculation)
from banana.exceptions import BananaUsageError
from banana.reference import FslReferenceData

logger = logging.getLogger('arcana')


class MriStudy(Study, metaclass=StudyMetaClass):

    add_data_specs = [
        InputFilesetSpec('magnitude', STD_IMAGE_FORMATS,
                         desc=("Typically the primary scan acquired from "
                               "the scanner for the given contrast")),
        InputFilesetSpec(
            'coreg_ref', STD_IMAGE_FORMATS,
            desc=("A reference scan to coregister the primary scan to. Should "
                  "not be brain extracted"),
            optional=True),
        InputFilesetSpec(
            'coreg_ref_brain', STD_IMAGE_FORMATS,
            desc=("A brain-extracted reference scan to coregister a brain-"
                  "extracted scan to. Note that the output of the "
                  "registration brain_coreg can also be derived by brain "
                  "extracting the output of coregistration performed "
                  "before brain extraction if 'coreg_ref' is provided"),
            optional=True),
        InputFilesetSpec(
            'channels', (multi_nifti_gz_format, zip_format),
            optional=True, desc=("Reconstructed complex image for each "
                                 "coil without standardisation.")),
        InputFilesetSpec('header_image', dicom_format, desc=(
            "A dataset that contains correct the header information for the "
            "acquired image. Used to copy geometry over preprocessed "
            "channels"), optional=True),
        FilesetSpec('mag_preproc', nifti_gz_format, 'prepare_pipeline',
                    desc=("Magnitude after basic preprocessing, such as "
                          "realigning image axis to a standard rotation")),
        FilesetSpec('mag_channels', multi_nifti_gz_format,
                    'preprocess_channels_pipeline'),
        FilesetSpec('phase_channels', multi_nifti_gz_format,
                    'preprocess_channels_pipeline'),
        FilesetSpec('brain', nifti_gz_format, 'brain_extraction_pipeline',
                    desc="The brain masked image"),
        FilesetSpec('brain_mask', nifti_gz_format, 'brain_extraction_pipeline',
                    desc="Mask of the brain"),
        FilesetSpec('mag_coreg', nifti_gz_format, 'coreg_pipeline',
                    desc="Head image coregistered to 'coreg_ref'"),
        FilesetSpec('brain_coreg', nifti_gz_format,
                    'brain_coreg_pipeline',
                    desc=("Either brain-extracted image coregistered to "
                          "'coreg_ref_brain' or a brain extraction of a "
                          "coregistered (incl. skull) image")),
        FilesetSpec('brain_mask_coreg', nifti_gz_format,
                    'brain_coreg_pipeline',
                    desc=("Either brain-extracted image coregistered to "
                          "'coreg_ref_brain' or a brain extraction of a "
                          "coregistered (incl. skull) image")),
        FilesetSpec('coreg_ants_mat', text_matrix_format,
                    'coreg_ants_mat_pipeline'),
        FilesetSpec('coreg_fsl_mat', text_matrix_format,
                    'coreg_fsl_mat_pipeline'),
        FilesetSpec('mag_coreg_to_tmpl', nifti_gz_format,
                    'coreg_to_tmpl_pipeline'),
        FilesetSpec('coreg_to_tmpl_fsl_coeff', nifti_gz_format,
                    'coreg_to_tmpl_pipeline'),
        FilesetSpec('coreg_to_tmpl_fsl_report', gif_format,
                    'coreg_to_tmpl_pipeline'),
        FilesetSpec('coreg_to_tmpl_ants_mat', text_matrix_format,
                    'coreg_to_tmpl_pipeline'),
        FilesetSpec('coreg_to_tmpl_ants_warp', nifti_gz_format,
                    'coreg_to_tmpl_pipeline'),
        FilesetSpec('motion_mats', motion_mats_format, 'motion_mat_pipeline'),
        FilesetSpec('qformed', nifti_gz_format, 'qform_transform_pipeline'),
        FilesetSpec('qform_mat', text_matrix_format,
                    'qform_transform_pipeline'),
        FieldSpec('tr', float, 'header_extraction_pipeline'),
        FieldSpec('echo_times', float, 'header_extraction_pipeline',
                  array=True),
        FieldSpec('voxel_sizes', float, 'header_extraction_pipeline',
                  array=True),
        FieldSpec('main_field_orient', float, 'header_extraction_pipeline',
                  array=True),
        FieldSpec('main_field_strength', float, 'header_extraction_pipeline'),
        FieldSpec('start_time', float, 'header_extraction_pipeline'),
        FieldSpec('real_duration', float, 'header_extraction_pipeline'),
        FieldSpec('total_duration', float, 'header_extraction_pipeline'),
        FieldSpec('ped', str, 'header_extraction_pipeline'),
        FieldSpec('pe_angle', float, 'header_extraction_pipeline'),
        # Templates
        InputFilesetSpec('template', STD_IMAGE_FORMATS, frequency='per_study',
                         default=FslReferenceData(
                             'MNI152_T1',
                             format=nifti_gz_format,
                             resolution='mni_template_resolution')),
        InputFilesetSpec('template_brain', STD_IMAGE_FORMATS,
                         frequency='per_study',
                         default=FslReferenceData(
                             'MNI152_T1',
                             format=nifti_gz_format,
                             resolution='mni_template_resolution',
                             dataset='brain')),
        InputFilesetSpec('template_mask', STD_IMAGE_FORMATS,
                         frequency='per_study',
                         default=FslReferenceData(
                             'MNI152_T1',
                             format=nifti_gz_format,
                             resolution='mni_template_resolution',
                             dataset='brain_mask'))]

    add_param_specs = [
        SwitchSpec('resample_coreg_ref', False,
                   desc=("Whether to resample the coregistration reference "
                         "image to the resolution of the moving image")),
        SwitchSpec('reorient_to_std', True),
        ParamSpec('force_channel_flip', None, dtype=str, array=True,
                      desc=("Forcibly flip channel inputs during preprocess "
                            "channels to correct issues with channel recon. "
                            "The inputs are passed directly through to FSL's "
                            "swapdims (see fsl.SwapDimensions interface)")),
        SwitchSpec('bet_robust', True),
        ParamSpec('bet_f_threshold', 0.5),
        SwitchSpec('bet_reduce_bias', False,
                   desc="Only used if not 'bet_robust'"),
        ParamSpec('bet_g_threshold', 0.0),
        SwitchSpec('bet_method', 'fsl_bet', ('fsl_bet', 'optibet')),
        SwitchSpec('optibet_gen_report', False),
        SwitchSpec('coreg_to_tmpl_method', 'ants', ('fnirt', 'ants')),
        ParamSpec('mni_template_resolution', None, choices=(0.5, 1, 2),
                  dtype=int),
        ParamSpec('fnirt_intensity_model', 'global_non_linear_with_bias'),
        ParamSpec('fnirt_subsampling', [4, 4, 2, 2, 1, 1]),
        ParamSpec('reoriented_dims', ('RL', 'AP', 'IS')),
        ParamSpec('resampled_resolution', None, dtype=list),
        SwitchSpec('coreg_method', 'ants', ('ants', 'flirt', 'spm'),
                   desc="The tool to use for linear registration"),
        ParamSpec('flirt_degrees_of_freedom', 6, desc=(
            "Number of degrees of freedom used in the registration. "
            "Default is 6 -> affine transformation.")),
        ParamSpec('flirt_cost_func', 'normmi', desc=(
            "Cost function used for the registration. Can be one of "
            "'mutualinfo', 'corratio', 'normcorr', 'normmi', 'leastsq',"
            " 'labeldiff', 'bbr'")),
        ParamSpec('flirt_qsform', False, desc=(
            "Whether to use the QS form supplied in the input image "
            "header (the image coordinates of the FOV supplied by the "
            "scanner")),
        ParamSpec(
            'channel_fname_regex',
            r'.*_(?P<channel>\d+)_(?P<echo>\d+)_(?P<axis>[A-Z]+)\.nii\.gz',
            desc=("The regular expression to extract channel, echo and complex"
                  " axis from the filenames of the coils channel images")),
        ParamSpec(
            'channel_real_label', 'REAL',
            desc=("The name of the real axis extracted from the channel "
                  "filename")),
        ParamSpec(
            'channel_imag_label', 'IMAGINARY',
            desc=("The name of the real axis extracted from the channel "
                  "filename"))]

    @property
    def mni_template_resolution(self):
        if self.parameter('mni_template_resolution') is not None:
            res = self.parameter('mni_template_resolution')
        else:
            raise ArcanaMissingDataException(
                "Automatic detection of dataset resolution is not implemented "
                "yet, please specify resolution of default MNI templates "
                "manually via 'mni_template_resolution' parameter")
        return res

    @property
    def is_coregistered(self):
        return self.provided('coreg_ref') or self.provided('coreg_ref_brain')

    @property
    def header_image_spec_name(self):
        if self.provided('header_image'):
            hdr_name = 'header_image'
        else:
            hdr_name = 'magnitude'
        return hdr_name

    @property
    def brain_spec_name(self):
        """
        The name of the brain extracted image after registration has been
        applied if registration is specified by supplying 'coreg_ref' or
        'coreg_ref_brain' optional inputs.
        """
        if self.is_coregistered:
            name = 'brain_coreg'
        else:
            name = 'brain'
        return name

    @property
    def brain_mask_spec_name(self):
        if self.is_coregistered:
            brain_mask = 'brain_mask_coreg'
        else:
            brain_mask = 'brain_mask'
        return brain_mask

    def preprocess_channels_pipeline(self, **name_maps):
        pipeline = self.new_pipeline(
            'preprocess_channels',
            name_maps=name_maps,
            desc=("Convert channel signals in complex coords to polar coords "
                  "and combine"))

        if (self.provided('header_image') or
                self.branch('reorient_to_std') or
                self.parameter('force_channel_flip') is not None):
            # Read channel files reorient them into standard space and then
            # write back to directory
            list_channels = pipeline.add(
                'list_channels',
                ListDir(),
                inputs={
                    'directory': ('channels', multi_nifti_gz_format)})

            if self.parameter('force_channel_flip') is not None:
                force_flip = pipeline.add(
                    'flip_dims',
                    fsl.SwapDimensions(
                        new_dims=tuple(self.parameter('force_channel_flip'))),
                    inputs={
                        'in_file': (list_channels, 'files')},
                    iterfield=['in_file'])
                geom_dest_file = (force_flip, 'out_file')
            else:
                geom_dest_file = (list_channels, 'files')

            if self.provided('header_image'):
                # If header image is provided stomp its geometry over the
                # acquired channels
                copy_geom = pipeline.add(
                    'qsm_copy_geometry',
                    fsl.CopyGeom(
                        output_type='NIFTI_GZ'),
                    inputs={
                        'in_file': ('header_image', nifti_gz_format),
                        'dest_file': geom_dest_file},
                    iterfield=(['dest_file']),
                    requirements=[fsl_req.v('5.0.8')])
                reorient_in_file = (copy_geom, 'out_file')
            else:
                reorient_in_file = geom_dest_file

            if self.branch('reorient_to_std'):
                reorient = pipeline.add(
                    'reorient_channel',
                    fsl.Reorient2Std(
                        output_type='NIFTI_GZ'),
                    inputs={
                        'in_file': reorient_in_file},
                    iterfield=['in_file'],
                    requirements=[fsl_req.v('5.0.8')])
                copy_to_dir_in_files = (reorient, 'out_file')
            else:
                copy_to_dir_in_files = reorient_in_file

            copy_to_dir = pipeline.add(
                'copy_to_dir',
                CopyToDir(),
                inputs={
                    'in_files': copy_to_dir_in_files,
                    'file_names': (list_channels, 'files')})
            to_polar_in_dir = (copy_to_dir, 'out_dir')
        else:
            to_polar_in_dir = ('channels', multi_nifti_gz_format)

        pipeline.add(
            'to_polar',
            ToPolarCoords(
                in_fname_re=self.parameter('channel_fname_regex'),
                real_label=self.parameter('channel_real_label'),
                imaginary_label=self.parameter('channel_imag_label')),
            inputs={
                'in_dir': to_polar_in_dir},
            outputs={
                'mag_channels': ('magnitudes_dir', multi_nifti_gz_format),
                'phase_channels': ('phases_dir', multi_nifti_gz_format)})

        return pipeline

    def coreg_pipeline(self, **name_maps):
        if self.branch('coreg_method', 'flirt'):
            pipeline = self._flirt_linear_coreg_pipeline(**name_maps)
        elif self.branch('coreg_method', 'ants'):
            pipeline = self._ants_linear_coreg_pipeline(**name_maps)
        elif self.branch('coreg_method', 'spm'):
            pipeline = self._spm_linear_coreg_pipeline(**name_maps)
        else:
            self.unhandled_branch('coreg_method')
        if not self.provided(pipeline.map_input('coreg_ref')):
            raise ArcanaOutputNotProducedException(
                "Cannot co-register {} as reference image "
                "'{}' has not been provided".format(
                    pipeline.map_input('coreg_ref')))
        return pipeline

    def brain_extraction_pipeline(self, **name_maps):
        if self.branch('bet_method', 'fsl_bet'):
            pipeline = self._bet_brain_extraction_pipeline(**name_maps)
        elif self.branch('bet_method', 'optibet'):
            pipeline = self._optiBET_brain_extraction_pipeline(**name_maps)
        else:
            self.unhandled_branch('bet_method')
        return pipeline

    def brain_coreg_pipeline(self, **name_maps):
        """
        Coregistered + brain-extracted images can be derived in 2-ways. If an
        explicit brain-extracted reference is provided to
        'coreg_ref_brain' then that is used to coregister a brain extracted
        image against. Alternatively, if only a skull-included reference is
        provided then the registration is performed with skulls-included and
        then brain extraction is performed after
        """
        if self.provided('coreg_ref_brain'):
            # If a reference brain extracted image is provided we coregister
            # the brain extracted image to that
            pipeline = self.coreg_pipeline(
                name='brain_coreg',
                name_maps=dict(
                    input_map={
                        'mag_preproc': 'brain',
                        'coreg_ref': 'coreg_ref_brain'},
                    output_map={
                        'mag_coreg': 'brain_coreg'},
                    name_maps=name_maps))

            # Apply coregistration transform to brain mask
            if self.branch('coreg_method', 'flirt'):
                pipeline.add(
                    'mask_transform',
                    ApplyXFM(
                        output_type='NIFTI_GZ',
                        apply_xfm=True),
                    inputs={
                        'in_matrix_file': (pipeline.node('flirt'),
                                           'out_matrix_file'),
                        'in_file': ('brain_mask', nifti_gz_format),
                        'reference': ('coreg_ref_brain', nifti_gz_format)},
                    outputs={
                        'brain_mask_coreg': ('out_file', nifti_gz_format)},
                    requirements=[fsl_req.v('5.0.10')],
                    wall_time=10)

            elif self.branch('coreg_method', 'ants'):
                # Convert ANTs transform matrix to FSL format if we have used
                # Ants registration so we can apply the transform using
                # ApplyXFM
                pipeline.add(
                    'mask_transform',
                    ants.resampling.ApplyTransforms(
                        interpolation='Linear',
                        input_image_type=3,
                        invert_transform_flags=[True, True, False]),
                    inputs={
                        'input_image': ('brain_mask', nifti_gz_format),
                        'reference_image': ('coreg_ref_brain',
                                            nifti_gz_format),
                        'transforms': (pipeline.node('ants_reg'),
                                       'forward_transforms')},
                    requirements=[ants_req.v('1.9')], mem_gb=16,
                    wall_time=30)
            else:
                self.unhandled_branch('coreg_method')

        elif self.provided('coreg_ref'):
            # If coreg_ref is provided then we co-register the non-brain
            # extracted images and then brain extract the co-registered image
            pipeline = self.brain_extraction_pipeline(
                name='bet_coreg',
                input_map={'mag_preproc': 'mag_coreg'},
                output_map={'brain': 'brain_coreg',
                            'brain_mask': 'brain_mask_coreg'},
                name_maps=name_maps)
        else:
            raise BananaUsageError(
                "Either 'coreg_ref' or 'coreg_ref_brain' needs to be provided "
                "in order to derive brain_coreg or brain_mask_coreg")
        return pipeline

    def _coreg_mat_pipeline(self, **name_maps):
        if self.provided('coreg_ref_brain'):
            pipeline = self.brain_coreg_pipeline(**name_maps)
        elif self.provided('coreg_ref'):
            pipeline = self.coreg_pipeline(**name_maps)
        else:
            raise ArcanaOutputNotProducedException(
                "'coregistration matrices can only be derived if 'coreg_ref' "
                "or 'coreg_ref_brain' is provided to {}".format(self))
        return pipeline

    def coreg_ants_mat_pipeline(self, **name_maps):
        if self.branch('coreg_method', 'ants'):
            pipeline = self._coreg_mat_pipeline(**name_maps)
        else:
            # Run the coreg_mat pipeline only to generate the ANTs transform
            # and mapping the typical outputs to None so they don't override
            # the other settings
            pipeline = self._coreg_mat_pipeline(
                output_maps={
                    'mag_preproc': None,
                    'brain_coreg': None,
                    'brain_mask_coreg': None},
                name_maps=name_maps)
        return pipeline

    def coreg_fsl_mat_pipeline(self, **name_maps):
        if self.branch('coreg_method', 'flirt'):
            pipeline = self._coreg_mat_pipeline(**name_maps)
        elif self.branch('coreg_method', 'ants'):
            # Convert ANTS transform to FSL transform
            pipeline = self.new_pipeline(
                name='convert_ants_to_fsl_coreg_mat',
                name_maps=name_maps)

            if self.provided('coreg_ref'):
                source = 'mag_preproc'
                ref = 'coreg_ref'
            elif self.provided('coreg_ref_brain'):
                source = 'brain'
                ref = 'coreg_ref_brain'
            else:
                raise BananaUsageError(
                    "Either 'coreg_ref' or 'coreg_ref_brain' needs to be "
                    "provided in order to derive brain_coreg or brain_coreg_"
                    "mask")

            pipeline.add(
                'transform_conv',
                ANTs2FSLMatrixConversion(
                    ras2fsl=True),
                inputs={
                    'itk_file': ('coreg_ants_mat', text_matrix_format),
                    'source_file': (source, nifti_gz_format),
                    'reference_file': (ref, nifti_gz_format)},
                outputs={
                    'coreg_fsl_mat': ('fsl_matrix', text_matrix_format)},
                requirements=[c3d_req.v('1.0')])
        else:
            self.unhandled_branch('coreg_method')

        return pipeline

    def coreg_to_tmpl_pipeline(self, **name_maps):
        if self.branch('coreg_to_tmpl_method', 'fnirt'):
            pipeline = self._fnirt_to_tmpl_pipeline(**name_maps)
        elif self.branch('coreg_to_tmpl_method', 'ants'):
            pipeline = self._ants_to_tmpl_pipeline(**name_maps)
        else:
            self.unhandled_branch('coreg_to_tmpl_method')
        return pipeline

    def _flirt_linear_coreg_pipeline(self, **name_maps):
        """
        Registers a MR scan to a refernce MR scan using FSL's FLIRT command
        """

        pipeline = self.new_pipeline(
            name='linear_coreg',
            name_maps=name_maps,
            desc="Registers a MR scan against a reference image using FLIRT",
            citations=[fsl_cite])

        pipeline.add(
            'flirt',
            FLIRT(dof=self.parameter('flirt_degrees_of_freedom'),
                  cost=self.parameter('flirt_cost_func'),
                  cost_func=self.parameter('flirt_cost_func'),
                  output_type='NIFTI_GZ'),
            inputs={
                'in_file': ('mag_preproc', nifti_gz_format),
                'reference': ('coreg_ref', nifti_gz_format)},
            outputs={
                'mag_coreg': ('out_file', nifti_gz_format),
                'coreg_fsl_mat': ('out_matrix_file', text_matrix_format)},
            requirements=[fsl_req.v('5.0.8')],
            wall_time=5)

        return pipeline

    def _ants_linear_coreg_pipeline(self, **name_maps):
        """
        Registers a MR scan to a refernce MR scan using ANTS's linear_reg
        command
        """

        pipeline = self.new_pipeline(
            name='linear_coreg',
            name_maps=name_maps,
            desc="Registers a MR scan against a reference image using ANTs",
            citations=[ants_cite])

        pipeline.add(
            'ANTs_linear_Reg',
            AntsRegSyn(
                num_dimensions=3,
                transformation='r'),
            inputs={
                'ref_file': ('coreg_ref', nifti_gz_format),
                'input_file': ('mag_preproc', nifti_gz_format)},
            outputs={
                'mag_coreg': ('reg_file', nifti_gz_format),
                'coreg_ants_mat': ('regmat', text_matrix_format)},
            wall_time=10,
            requirements=[ants_req.v('2.0')])


#         ants_reg = pipeline.add(
#             'ants_reg',
#             ants.Registration(
#                 dimension=3,
#                 collapse_output_transforms=True,
#                 float=False,
#                 interpolation='Linear',
#                 use_histogram_matching=False,
#                 winsorize_upper_quantile=0.995,
#                 winsorize_lower_quantile=0.005,
#                 verbose=True,
#                 transforms=['Rigid'],
#                 transform_parameters=[(0.1,)],
#                 metric=['MI'],
#                 metric_weight=[1],
#                 radius_or_number_of_bins=[32],
#                 sampling_strategy=['Regular'],
#                 sampling_percentage=[0.25],
#                 number_of_iterations=[[1000, 500, 250, 100]],
#                 convergence_threshold=[1e-6],
#                 convergence_window_size=[10],
#                 shrink_factors=[[8, 4, 2, 1]],
#                 smoothing_sigmas=[[3, 2, 1, 0]],
#                 output_warped_image=True),
#             inputs={
#                 'fixed_image': ('coreg_ref', nifti_gz_format),
#                 'moving_image': ('mag_preproc', nifti_gz_format)},
#             outputs={
#                 'mag_coreg': ('warped_image', nifti_gz_format)},
#             wall_time=10,
#             requirements=[ants_req.v('2.0')])
#
#         pipeline.add(
#             'select',
#             SelectOne(
#                 index=0),
#             inputs={
#                 'inlist': (ants_reg, 'forward_transforms')},
#             outputs={
#                 'coreg_ants_mat': ('out', text_matrix_format)})

        return pipeline

    def _spm_linear_coreg_pipeline(self, **name_maps):  # @UnusedVariable
        """
        Coregisters T2 image to T1 image using SPM's "Register" method.

        NB: Default values come from the W2MHS toolbox
        """
        pipeline = self.new_pipeline(
            'linear_coreg',
            name_maps=name_maps,
            desc="Coregister T2-weighted images to T1",
            citations=[spm_cite])

        pipeline.add(
            'mag_coreg',
            Coregister(
                jobtype='estwrite',
                cost_function='nmi',
                separation=[4, 2],
                tolerance=[0.02, 0.02, 0.02, 0.001, 0.001, 0.001, 0.01, 0.01,
                           0.01, 0.001, 0.001, 0.001],
                fwhm=[7, 7],
                write_interp=4,
                write_wrap=[0, 0, 0],
                write_mask=False,
                out_prefix='r'),
            inputs={
                'target': ('coreg_ref', nifti_format),
                'source': ('mag_preproc', nifti_format)},
            outputs={
                'mag_coreg': ('coregistered_source', nifti_format)},
            requirements=[spm_req.v(12)],
            wall_time=30)
        return pipeline

    def qform_transform_pipeline(self, **name_maps):
        pipeline = self.new_pipeline(
            name='qform_transform',
            name_maps=name_maps,
            desc="Registers a MR scan against a reference image",
            citations=[fsl_cite])

        if self.provided('coreg_ref'):
            in_file = 'mag_preproc'
            reference = 'coreg_ref'
        elif self.provided('coreg_ref_brain'):
            in_file = 'brain'
            reference = 'coreg_ref_brain'
        else:
            raise BananaUsageError(
                "'coreg_ref' or 'coreg_ref_brain' need to be provided to "
                "study in order to run qform_transform")

        pipeline.add(
            'flirt',
            FLIRT(
                uses_qform=True,
                apply_xfm=True,
                output_type='NIFTI_GZ'),
            inputs={
                'in_file': (in_file, nifti_gz_format),
                'reference': (reference, nifti_gz_format)},
            outputs={
                'qformed': ('out_file', nifti_gz_format),
                'qform_mat': ('out_matrix_file', text_matrix_format)},
            requirements=[fsl_req.v('5.0.8')],
            wall_time=5)

        return pipeline

    def _bet_brain_extraction_pipeline(self, **name_maps):
        """
        Generates a whole brain mask using FSL's BET command.
        """
        pipeline = self.new_pipeline(
            name='brain_extraction',
            name_maps=name_maps,
            desc="Generate brain mask from mr_scan",
            citations=[fsl_cite, bet_cite, bet2_cite])
        # Create mask node
        bet = pipeline.add(
            "bet",
            fsl.BET(
                mask=True,
                output_type='NIFTI_GZ',
                frac=self.parameter('bet_f_threshold'),
                vertical_gradient=self.parameter('bet_g_threshold')),
            inputs={
                'in_file': ('mag_preproc', nifti_gz_format)},
            outputs={
                'brain': ('out_file', nifti_gz_format),
                'brain_mask': ('mask_file', nifti_gz_format)},
            requirements=[fsl_req.v('5.0.9')])
        # Set either robust or reduce bias
        if self.branch('bet_robust'):
            bet.inputs.robust = True
        else:
            bet.inputs.reduce_bias = self.parameter('bet_reduce_bias')
        return pipeline

    def _optiBET_brain_extraction_pipeline(self, **name_maps):
        """
        Generates a whole brain mask using a modified optiBET approach.
        """
        pipeline = self.new_pipeline(
            name='brain_extraction',
            name_maps=name_maps,
            desc=("Modified implementation of optiBET.sh"),
            citations=[fsl_cite])

        mni_reg = pipeline.add(
            'T1_reg',
            AntsRegSyn(
                num_dimensions=3,
                transformation='s',
                out_prefix='T12MNI',
                num_threads=4),
            inputs={
                'ref_file': ('template', nifti_gz_format),
                'input_file': ('mag_preproc', nifti_gz_format)},
            wall_time=25,
            requirements=[ants_req.v('2.0')])

        merge_trans = pipeline.add(
            'merge_transforms',
            Merge(2),
            inputs={
                'in1': (mni_reg, 'inv_warp'),
                'in2': (mni_reg, 'regmat')},
            wall_time=1)

        trans_flags = pipeline.add(
            'trans_flags',
            Merge(2,
                  in1=False,
                  in2=True),
            wall_time=1)

        apply_trans = pipeline.add(
            'ApplyTransform',
            ApplyTransforms(
                interpolation='NearestNeighbor',
                input_image_type=3),
            inputs={
                'input_image': ('template_mask', nifti_gz_format),
                'reference_image': ('mag_preproc', nifti_gz_format),
                'transforms': (merge_trans, 'out'),
                'invert_transform_flags': (trans_flags, 'out')},
            wall_time=7,
            mem_gb=24,
            requirements=[ants_req.v('2.0')])

        maths1 = pipeline.add(
            'binarize',
            fsl.ImageMaths(
                suffix='_optiBET_brain_mask',
                op_string='-bin',
                output_type='NIFTI_GZ'),
            inputs={
                'in_file': (apply_trans, 'output_image')},
            outputs={
                'brain_mask': ('out_file', nifti_gz_format)},
            wall_time=5,
            requirements=[fsl_req.v('5.0.8')])

        maths2 = pipeline.add(
            'mask',
            fsl.ImageMaths(
                suffix='_optiBET_brain',
                op_string='-mas',
                output_type='NIFTI_GZ'),
            inputs={
                'in_file': ('mag_preproc', nifti_gz_format),
                'in_file2': (maths1, 'out_file')},
            outputs={
                'brain': ('out_file', nifti_gz_format)},
            wall_time=5,
            requirements=[fsl_req.v('5.0.8')])

        if self.branch('optibet_gen_report'):
            pipeline.add(
                'slices',
                FSLSlices(
                    outname='optiBET_report',
                    output_type='NIFTI_GZ'),
                wall_time=5,
                inputs={
                    'im1': ('mag_preproc', nifti_gz_format),
                    'im2': (maths2, 'out_file')},
                outputs={
                    'optiBET_report': ('report', gif_format)},
                requirements=[fsl_req.v('5.0.8')])

        return pipeline

    # @UnusedVariable @IgnorePep8
    def _fnirt_to_tmpl_pipeline(self, **name_maps):
        """
        Registers a MR scan to a refernce MR scan using FSL's nonlinear FNIRT
        command

        Parameters
        ----------
        template : Which template to use, can be one of 'mni_nl6'
        """
        pipeline = self.new_pipeline(
            name='mag_coreg_to_tmpl',
            name_maps=name_maps,
            desc=("Nonlinearly registers a MR scan to a standard space,"
                  "e.g. MNI-space"),
            citations=[fsl_cite])

        # Basic reorientation to standard MNI space
        reorient = pipeline.add(
            'reorient',
            Reorient2Std(
                output_type='NIFTI_GZ'),
            inputs={
                'in_file': ('mag_preproc', nifti_gz_format)},
            requirements=[fsl_req.v('5.0.8')])

        reorient_mask = pipeline.add(
            'reorient_mask',
            Reorient2Std(
                output_type='NIFTI_GZ'),
            inputs={
                'in_file': ('brain_mask', nifti_gz_format)},
            requirements=[fsl_req.v('5.0.8')])

        reorient_brain = pipeline.add(
            'reorient_brain',
            Reorient2Std(
                output_type='NIFTI_GZ'),
            inputs={
                'in_file': ('brain', nifti_gz_format)},
            requirements=[fsl_req.v('5.0.8')])

        # Affine transformation to MNI space
        flirt = pipeline.add(
            'flirt',
            interface=FLIRT(
                dof=12,
                output_type='NIFTI_GZ'),
            inputs={
                'reference': ('template_brain', nifti_gz_format),
                'in_file': (reorient_brain, 'out_file')},
            requirements=[fsl_req.v('5.0.8')],
            wall_time=5)

        # Apply mask if corresponding subsampling scheme is 1
        # (i.e. 1-to-1 resolution) otherwise don't.
        apply_mask = [int(s == 1)
                      for s in self.parameter('fnirt_subsampling')]
        # Nonlinear transformation to MNI space
        pipeline.add(
            'fnirt',
            interface=FNIRT(
                output_type='NIFTI_GZ',
                intensity_mapping_model=(
                    self.parameter('fnirt_intensity_model')
                    if self.parameter('fnirt_intensity_model') is not None else
                    'none'),
                subsampling_scheme=self.parameter('fnirt_subsampling'),
                fieldcoeff_file=True,
                in_fwhm=[8, 6, 5, 4, 3, 2],  # [8, 6, 5, 4.5, 3, 2] This threw an error because of float value @IgnorePep8,
                ref_fwhm=[8, 6, 5, 4, 2, 0],
                regularization_lambda=[300, 150, 100, 50, 40, 30],
                apply_intensity_mapping=[1, 1, 1, 1, 1, 0],
                max_nonlin_iter=[5, 5, 5, 5, 5, 10],
                apply_inmask=apply_mask,
                apply_refmask=apply_mask),
            inputs={
                'ref_file': ('template', nifti_gz_format),
                'refmask': ('template_mask', nifti_gz_format),
                'in_file': (reorient, 'out_file'),
                'inmask_file': (reorient_mask, 'out_file'),
                'affine_file': (flirt, 'out_matrix_file')},
            outputs={
                'mag_coreg_to_tmpl': ('warped_file', nifti_gz_format),
                'coreg_to_tmpl_fsl_coeff': ('fieldcoeff_file',
                                             nifti_gz_format)},
            requirements=[fsl_req.v('5.0.8')],
            wall_time=60)
        # Set registration parameters
        # TODO: Need to work out which parameters to use
        return pipeline

    def _ants_to_tmpl_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='mag_coreg_to_tmpl',
            name_maps=name_maps,
            desc=("Nonlinearly registers a MR scan to a standard space,"
                  "e.g. MNI-space"),
            citations=[fsl_cite])

        pipeline.add(
            'Struct2MNI_reg',
            AntsRegSyn(
                num_dimensions=3,
                transformation='s',
                num_threads=4),
            inputs={
                'input_file': (self.brain_spec_name, nifti_gz_format),
                'ref_file': ('template_brain', nifti_gz_format)},
            outputs={
                'mag_coreg_to_tmpl': ('reg_file', nifti_gz_format),
                'coreg_to_tmpl_ants_mat': ('regmat', text_matrix_format),
                'coreg_to_tmpl_ants_warp': ('warp_file', nifti_gz_format)},
            wall_time=25,
            requirements=[ants_req.v('2.0')])

#         ants_reg = pipeline.add(
#             'ants_reg',
#             ants.Registration(
#                 dimension=3,
#                 collapse_output_transforms=True,
#                 float=False,
#                 interpolation='Linear',
#                 use_histogram_matching=False,
#                 winsorize_upper_quantile=0.995,
#                 winsorize_lower_quantile=0.005,
#                 verbose=True,
#                 transforms=['Rigid', 'Affine', 'SyN'],
#                 transform_parameters=[(0.1,), (0.1,), (0.1, 3, 0)],
#                 metric=['MI', 'MI', 'CC'],
#                 metric_weight=[1, 1, 1],
#                 radius_or_number_of_bins=[32, 32, 32],
#                 sampling_strategy=['Regular', 'Regular', 'None'],
#                 sampling_percentage=[0.25, 0.25, None],
#                 number_of_iterations=[[1000, 500, 250, 100],
#                                       [1000, 500, 250, 100],
#                                       [100, 70, 50, 20]],
#                 convergence_threshold=[1e-6, 1e-6, 1e-6],
#                 convergence_window_size=[10, 10, 10],
#                 shrink_factors=[[8, 4, 2, 1],
#                                 [8, 4, 2, 1],
#                                 [8, 4, 2, 1]],
#                 smoothing_sigmas=[[3, 2, 1, 0],
#                                   [3, 2, 1, 0],
#                                   [3, 2, 1, 0]],
#                 output_warped_image=True),
#             inputs={
#                 'fixed_image': ('template_brain', nifti_gz_format),
#                 'moving_image': (self.brain_spec_name, nifti_gz_format)},
#             outputs={
#                 'mag_coreg_to_tmpl': ('warped_image', nifti_gz_format)},
#             wall_time=25,
#             requirements=[ants_req.v('2.0')])
#
#         select_trans = pipeline.add(
#             'select',
#             SelectOne(
#                 index=1),
#             inputs={
#                 'inlist': (ants_reg, 'forward_transforms')},
#             outputs={
#                 'coreg_to_tmpl_ants_mat': ('out', text_matrix_format)})
#
#         pipeline.add(
#             'select_warp',
#             SelectOne(
#                 index=0),
#             inputs={
#                 'inlist': (ants_reg, 'forward_transforms')},
#             outputs={
#                 'coreg_to_tmpl_ants_warp': ('out', nifti_gz_format)})
#
#         pipeline.add(
#             'slices',
#             FSLSlices(
#                 outname='coreg_to_tmpl_report'),
#             inputs={
#                 'im1': ('template', nifti_gz_format),
#                 'im2': (select_trans, 'out')},
#             outputs={
#                 'coreg_to_tmpl_fsl_report': ('report', gif_format)},
#             wall_time=1,
#             requirements=[fsl_req.v('5.0.8')])

        return pipeline

    def prepare_pipeline(self, **name_maps):
        """
        Performs basic preprocessing, such as swapping dimensions into
        standard orientation and resampling (if required)

        Parameters
        -------
        new_dims : tuple(str)[3]
            A 3-tuple with the new orientation of the image (see FSL
            swap dim)
        resolution : list(float)[3] | None
            New resolution of the image. If None no resampling is
            performed
        """
        pipeline = self.new_pipeline(
            name='prepare_pipeline',
            name_maps=name_maps,
            desc=("Dimensions swapping to ensure that all the images "
                  "have the same orientations."),
            citations=[fsl_cite])

        if (self.branch('reorient_to_std') or
                self.parameter('resampled_resolution') is not None):
            if self.branch('reorient_to_std'):
                swap = pipeline.add(
                    'fslreorient2std',
                    fsl.utils.Reorient2Std(
                        output_type='NIFTI_GZ'),
                    inputs={
                        'in_file': ('magnitude', nifti_gz_format)},
                    requirements=[fsl_req.v('5.0.9')])
    #         swap.inputs.new_dims = self.parameter('reoriented_dims')

            if self.parameter('resampled_resolution') is not None:
                resample = pipeline.add(
                    "resample",
                    MRResize(
                        voxel=self.parameter('resampled_resolution')),
                    inputs={
                        'in_file': (swap, 'out_file')},
                    requirements=[mrtrix_req.v('3.0rc3')])
                pipeline.connect_output('mag_preproc', resample, 'out_file',
                                        nifti_gz_format)
            else:
                pipeline.connect_output('mag_preproc', swap, 'out_file',
                                        nifti_gz_format)
        else:
            # Don't actually do any processing just copy magnitude image to
            # preproc
            pipeline.add(
                'identity',
                IdentityInterface(
                    ['file']),
                inputs={
                    'file': ('magnitude', nifti_gz_format)},
                outputs={
                    'mag_preproc': ('file', nifti_gz_format)})

        return pipeline

    def header_extraction_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='header_extraction',
            name_maps=name_maps,
            desc=("Pipeline to extract the most important scan "
                  "information from the image header"),
            citations=[])

        input_format = self.input(self.header_image_spec_name).format

        if input_format == dicom_format:

            pipeline.add(
                'hd_info_extraction',
                DicomHeaderInfoExtraction(
                    multivol=False),
                inputs={
                    'dicom_folder': (self.header_image_spec_name, dicom_format)},
                outputs={
                    'tr': ('tr', float),
                    'start_time': ('start_time', str),
                    'total_duration': ('total_duration', str),
                    'real_duration': ('real_duration', str),
                    'ped': ('ped', str),
                    'pe_angle': ('pe_angle', str),
                    'echo_times': ('echo_times', float),
                    'voxel_sizes': ('voxel_sizes', float),
                    'main_field_strength': ('B0', float),
                    'main_field_orient': ('H', float)})

        elif input_format == nifti_gz_x_format:

            pipeline.add(
                'hd_info_extraction',
                NiftixHeaderInfoExtraction(),
                inputs={
                    'in_file': (self.header_image_spec_name, nifti_gz_x_format)},
                outputs={
                    'tr': ('tr', float),
                    'start_time': ('start_time', str),
                    'total_duration': ('total_duration', str),
                    'real_duration': ('real_duration', str),
                    'ped': ('ped', str),
                    'pe_angle': ('pe_angle', str),
                    'echo_times': ('echo_times', float),
                    'voxel_sizes': ('voxel_sizes', float),
                    'main_field_strength': ('B0', float),
                    'main_field_orient': ('H', float)})
        else:
            raise BananaUsageError(
                "Can only extract header info if 'magnitude' fileset "
                "is provided in DICOM or extended NIfTI format (provided {})"
                .format(self.input(self.header_image_spec_name).format))

        return pipeline

    def motion_mat_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='motion_mat_calculation',
            name_maps=name_maps,
            desc=("Motion matrices calculation"),
            citations=[fsl_cite])

        mm = pipeline.add(
            'motion_mats',
            MotionMatCalculation(),
            outputs={
                'motion_mats': ('motion_mats', motion_mats_format)})
        if not self.spec('coreg_fsl_mat').derivable:
            logger.info("Cannot derive 'coreg_matrix' for {} required for "
                        "motion matrix calculation, assuming that it "
                        "is the reference study".format(self))
            mm.inputs.reference = True
            pipeline.connect_input('magnitude', mm, 'dummy_input')
        else:
            pipeline.connect_input('coreg_fsl_mat', mm, 'reg_mat',
                                   text_matrix_format)
            pipeline.connect_input('qform_mat', mm, 'qform_mat',
                                   text_matrix_format)
            if 'align_mats' in self.data_spec_names():
                pipeline.connect_input('align_mats', mm, 'align_mats',
                                       motion_mats_format)
        return pipeline



import os.path as op
from nipype.interfaces.base import (
    BaseInterface, BaseInterfaceInputSpec, traits, File, TraitedSpec,
    Directory, isdefined)
from nipype.interfaces.freesurfer.base import FSCommand, FSTraitedSpec


class GenFSBrainMasksInputSpec(BaseInterfaceInputSpec):

    fs_directory = Directory(exists=True, mandatory=True, desc='Directory with'
                             'freesurfer results.')


class GenFSBrainMasksOutputSpec(TraitedSpec):

    totGM = File(exists=True, desc='Total grey matter mask')
    corticalGM = File(exists=True, desc='Cortical grey matter mask')
    totWM = File(exists=True, desc='Total white matter mask')
    subcorticalGM = File(exists=True, desc='Subcortical grey matter mask')
    totCSF = File(exists=True, desc='Total CSF mask')
    totBrainVol = File(exists=True, desc='Total brain volume mask')
    totBrainVolNoVent = File(exists=True, desc='Total brain volume mask'
                             'without ventricles')
    Brainstem = File(exists=True, desc='Brainstem mask')


class GenFSBrainMasks(BaseInterface):

    input_spec = GenFSBrainMasksInputSpec
    output_spec = GenFSBrainMasksOutputSpec

    def _run_interface(self, runtime):
        pass


class AparcStatsInputSpec(FSTraitedSpec):

    subjects = traits.List(
        traits.Str,
        argstr='--subjects %s',
        desc="Directories containing the recon-all outputs to summarise")

    tablefile = File(argstr='--tablefile %s', genfile=True,
                     desc=("(REQUIRED) output table file"))
    hemisphere = traits.Enum(
        'lh', 'rh',
        argstr='--hemi %s',
        mandatory=True,
        desc=("(REQUIRED) lh or rh"))
    qdec = traits.Str('fsid', argstr='--qdec %s',
                      desc=("name of the qdec table which has the column of "
                            "subjects ids (fsid)"))
    qdec_long = traits.Str(
        argstr='--qdec-long %s',
        desc=("name of the longitudinal qdec table which has the column of tp "
              "ids (fsid) and subject templates (fsid-base)"))
    parc = traits.Enum(
        'aparc', 'aparc.a2009s', 'aparc.DKTatlas40',
        argstr='--parc %s', usedefault=True,
        desc=("parcellation.. default is aparc (alt aparc.a2009s)"))
    measure = traits.Enum(
        'volume', 'thickness', 'thicknessstd', 'meancurv', 'gauscurv',
        'foldind', 'curvind',
        argstr='--measure %s',
        desc=("measure: default is area ()"))
    delimiter = traits.Enum(
        'tab', 'comma', 'space', 'semicolon',
        argstr='--delimiter %s',
        desc=("delimiter between measures in the table. "
              "default is tab (alt comma, space, semicolon, tab)"))
    skip = traits.Any(argstr='--skip %s',
                      desc=("if a subject does not have input, skip it"))
    parcid_only = traits.Bool(
        argstr='--parcid-only',
        desc=("do not pre/append hemi/meas to parcellation name"))
    common_parcs = traits.Bool(
        argstr='--common-parcs',
        desc=("output only the common parcellations of all the subjects "
              "given"))
    parcs_from_file = traits.Str(
        argstr='--parcs-from-file %s',
        desc=("filename: output parcellations specified in the file"))
    report_rois = traits.Bool(
        argstr='--report-rois',
        desc=("print ROIs information for each subject"))
    transpose = traits.Bool(
        argstr='--transpose',
        desc=("transpose the table ( default is subjects in rows and ROIs in "
              "cols)"))


class AparcStatsOutputSpec(TraitedSpec):

    tablefile = File(exists=True, desc="The output table file")


class AparcStats(FSCommand):

    input_spec = AparcStatsInputSpec
    output_spec = AparcStatsOutputSpec
    _cmd = 'aparcstats2table'

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['tablefile'] = self._gen_tablefile_fname()
        return outputs

    def _gen_filename(self, name):
        if name == 'tablefile':
            fname = self._gen_tablefile_fname()
        else:
            assert False
        return fname

    def _gen_tablefile_fname(self):
        if isdefined(self.inputs.tablefile):
            fname = self.inputs.tablefile
        else:
            fname = 'aparc_{}_{}_table.txt'.format(self.inputs.hemisphere,
                                                   self.inputs.measure)
        return op.abspath(fname)



import os.path
from nipype.interfaces.base import (
    BaseInterface, File, TraitedSpec, traits, isdefined,
    BaseInterfaceInputSpec)
from nipype.interfaces.matlab import MatlabCommand
from arcana.exceptions import ArcanaError
from arcana.utils import split_extension


class CreateROIInputSpec(BaseInterfaceInputSpec):

    in_file = traits.File(  # @UndefinedVariable
        exists=True, desc="Input diffusion file to create the ROI for",
        argstr='%s', position=0, mandatory=True)

    brain_mask = traits.File(  # @UndefinedVariable
        exists=True, desc="Whole brain mask", argstr='%s', position=1,
        mandatory=True)

    out_file = traits.File(  # @UndefinedVariable
        genfile=True, argstr="%s", hash_files=False, position=2,
        desc="The name of the ROI file to be generated")


class CreateROIOutputSpec(TraitedSpec):

    out_file = File(exists=True, desc='ROI for NODDI processing')


class CreateROI(BaseInterface):

    input_spec = CreateROIInputSpec
    output_spec = CreateROIInputSpec

    def _run_interface(self, runtime):  # @UnusedVariable
        script = "CreateROI('{}', '{}', '{}');".format(
            self.inputs.in_file, self.inputs.brain_mask,
            self._gen_outfilename())
        mlab = MatlabCommand(script=script, mfile=True)
        result = mlab.run()
        return result.runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self._gen_outfilename()
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            gen_name = self._gen_outfilename()
        else:
            assert False
        return gen_name

    def _gen_outfilename(self):
        if isdefined(self.inputs.out_file):
            if not self.inputs.out_file.endswith('.mat'):
                raise ArcanaError(
                    "Output NODDI ROI should be saved with '.mat' extension "
                    "(provided '{}')".format(self.inputs.out_file))
            out_name = self.inputs.out_file
        else:
            base, _ = split_extension(os.path.basename(self.inputs.in_file))
            out_name = os.path.join(os.getcwd(), "{}_ROI.mat".format(base))
        return out_name


class BatchNODDIFittingInputSpec(BaseInterfaceInputSpec):

    roi_file = traits.File(  # @UndefinedVariable
        exists=True, mandatory=True,
        desc="Input ROI to fit the parameters for")

    bvecs_file = File(
        exists=True, mandatory=True,
        desc=("Gradient encoding directions in FSL format"))

    bvals_file = File(
        exists=True, mandatory=True,
        desc="Extracted graident encoding b-values in FSL format")

    model = traits.Str(  # @UndefinedVariable
        'WatsonSHStickTortIsoV_B0', mandatory=False, usedefault=True,
        desc="The NODDI model used (see NODDI docs for alternatives)")

    nthreads = traits.Int(  # @UndefinedVariable
        4, mandatory=False, usedefault=True,
        desc="The number of compute cores to use for the fitting process")

    out_file = traits.File(  # @UndefinedVariable
        genfile=True, hash_files=False,
        desc="The name of the ROI file to be generated")


class BatchNODDIFittingOutputSpec(TraitedSpec):

    out_file = File(exists=True, desc='ROI for NODDI processing')


class BatchNODDIFitting(BaseInterface):

    input_spec = BatchNODDIFittingInputSpec
    output_spec = BatchNODDIFittingOutputSpec

    def _run_interface(self, runtime):  # @UnusedVariable
        script = """
        protocol = FSL2Protocol('{bvals}', '{bvecs}');
        noddi = MakeModel('{model}');
        batch_fitting('{roi}', protocol, noddi, '{out_file}', {nthreads});
        """.format(
            bvecs=self.inputs.bvecs_file, bvals=self.inputs.bvals_file,
            model=self.inputs.model, roi=self.inputs.roi_file,
            out_file=self._gen_outfilename(), nthreads=self.inputs.nthreads)
        mlab = MatlabCommand(script=script, mfile=True)
        result = mlab.run()
        return result.runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self._gen_outfilename()
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            gen_name = self._gen_outfilename()
        else:
            assert False
        return gen_name

    def _gen_outfilename(self):
        if isdefined(self.inputs.out_file):
            out_name = self.inputs.out_file
        else:
            base, _ = split_extension(os.path.basename(self.inputs.roi_file))
            out_name = os.path.join(
                os.getcwd(), "{}_fitted_params.mat".format(base))
        return out_name


class SaveParamsAsNIfTIInputSpec(BaseInterfaceInputSpec):

    params_file = File(
        exists=True, mandatory=True,
        desc="The parameters fitted by BatchNODDIFitting")

    roi_file = File(
        exists=True, mandatory=True, desc="The ROI file created by CreateROI")

    brain_mask_file = File(
        exists=True, mandatory=True, desc="A whole brain mask")

    output_prefix = traits.Str(  # @UndefinedVariable
        "processed_noddi", mandatory=False,
        desc="Prefix of the generated output files")


class SaveParamsAsNIfTIOutputSpec(TraitedSpec):

    ficvf = File(
        exists=True,
        desc="Neurite density (or intra-cellular volume fraction)")

    odi = File(exists=True, desc="Orientation dispersion index (ODI)")

    fiso = File(exists=True, desc="CSF volume fraction")

    fibredirs_xvec = File(exists=True, desc="X fibre orientation")

    fibredirs_yvec = File(exists=True, desc="Y fibre orientation")

    fibredirs_zvec = File(exists=True, desc="Z fibre orientation")

    fmin = File(exists=True, desc="Fitting objective function values")

    kappa = File(
        exists=True,
        desc=("Concentration parameter of Watson distribution used to "
              "compute ODI"))

    error_code = File(
        exists=True,
        desc="Error code (NEW) Nonzero values indicate fitting errors")


class SaveParamsAsNIfTI(BaseInterface):

    input_spec = SaveParamsAsNIfTIInputSpec
    output_spec = SaveParamsAsNIfTIOutputSpec

    def _run_interface(self, runtime):  # @UnusedVariable
        script = """
        SaveParamsAsNIfTI('{params}', '{roi}', '{brain_mask}', '{prefix}');
        """.format(
            params=self.inputs.params_file, roi=self.inputs.roi_file,
            brain_mask=self.inputs.brain_mask_file,
            prefix=self.inputs.output_prefix)
        mlab = MatlabCommand(script=script, mfile=True)
        result = mlab.run()
        return result.runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        for name in ('ficvf', 'odi', 'fiso', 'fibredirs_xvec',
                     'fibredirs_yvec', 'fibredirs_zvec', 'fmin', 'kappa',
                     'error_code'):
            outputs[name] = os.path.join(
                os.getcwd(), '{}_{}.nii'.format(self.inputs.output_prefix,
                                                name))
        return outputs



from __future__ import absolute_import
import os.path as op
import numpy as np
from nipype.interfaces.base import (
    TraitedSpec, BaseInterface, File, isdefined, traits)


class TransformGradientsInputSpec(TraitedSpec):
    gradients = File(exists=True, mandatory=True,
                     desc='input gradients to transform in FSL bvec format')
    transform = File(desc='The affine transform (output from FLIRT)')
    transformed = traits.Str(desc='the name for the transformed file')


class TransformGradientsOutputSpec(TraitedSpec):
    transformed = File(exists=True, desc='the transformed gradients')


class TransformGradients(BaseInterface):
    """
    Applies a coregistration transform matrix (in FLIRT format) to a FSL bvec
    file
    """

    input_spec = TransformGradientsInputSpec
    output_spec = TransformGradientsOutputSpec

    def _run_interface(self, runtime):
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        rotation = np.loadtxt(self.inputs.transform)[:3, :3]
        gradients = np.loadtxt(self.inputs.gradients)
        transformed = rotation.dot(gradients)
        np.savetxt(self.transformed_path, transformed)
        outputs['transformed'] = self.transformed_path
        return outputs

    @property
    def transformed_path(self):
        if isdefined(self.inputs.transformed):
            dpath = self.inputs.transformed
        else:
            dpath = op.abspath('transformed')
        return dpath



import sys
import os.path as op
import os
from argparse import ArgumentParser
from importlib import import_module
from banana.utils.testing import PipelineTester
from banana.exceptions import BananaUsageError
from multiprocessing import cpu_count
from arcana.utils import parse_value
from banana import (
    InputFilesets, MultiProc, SingleProc, SlurmProc, StaticEnv, ModulesEnv,
    BasicRepo, BidsRepo, XnatRepo)
import logging

logger = logging.getLogger('banana')


def set_loggers(loggers):

    # Overwrite earlier (default) versions of logger levels with later options
    loggers = dict(loggers)

    for name, level in loggers.items():
        logger = logging.getLogger(name)
        logger.setLevel(level)
        handler = logging.StreamHandler()
        formatter = logging.Formatter("%(levelname)s - %(message)s")
        handler.setFormatter(formatter)
        logger.addHandler(handler)


def resolve_class(class_str, prefixes=('banana.', 'banana.study.')):
    parts = class_str.split('.')
    module_name = '.'.join(parts[:-1])
    class_name = parts[-1]
    cls = None
    for prefix in [''] + list(prefixes):
        mod_name = prefix + module_name
        if not mod_name:
            continue
        mod_name = mod_name.strip('.')
        try:
            module = import_module(mod_name)
        except ModuleNotFoundError:
            continue
        else:
            try:
                cls = getattr(module, class_name)
            except AttributeError:
                continue
            else:
                break
    if cls is None:
        raise BananaUsageError(
            "Did not find class '{}'".format(class_str))
    return cls


class DeriveCmd():

    desc = "Generate derivatives from a Banana Study class"

    @classmethod
    def parser(cls):
        parser = ArgumentParser(prog='banana derive',
                                description=cls.desc)
        parser.add_argument('repository_path',
                            help=("Either the path to the repository if of "
                                  "'bids' or 'basic' types, or the name of the"
                                  "project ID for 'xnat' type"))
        parser.add_argument('study_class',
                            help=("Name of the class to instantiate"))
        parser.add_argument('study_name',
                            help=("The name of the study to put the analysis "
                                  "under (e.g. parenthood)"))
        parser.add_argument('derivatives', nargs='+', metavar=['SPEC'],
                            help=("The names of the derivatives to generate"))
        parser.add_argument('--repository', nargs='+', default=['bids'],
                            metavar=['TYPE', 'REPO_ARGS'],
                            help=("Specify the repository type and any options"
                                  " to be passed to it. First argument "))
        parser.add_argument('--processor', default=['multi'], nargs='+',
                            metavar=['TYPE', 'PROC_ARGS'],
                            help=("The type of processor to use plus arguments"
                                  "used to initate it. First arg is the type "
                                  "(one of 'single', 'multi', 'slurm'). "
                                  "Additional arguments depend on type: "
                                  "single [], multi [NUM_PROCS], slurm ["
                                  "ACCOUNT, PARTITION]"))
        parser.add_argument('--environment', type=str, default='static',
                            choices=('modules', 'static'), metavar=['TYPE'],
                            help="The type of environment to use")
        parser.add_argument('--input', '-i', nargs=2, action='append',
                            default=[], metavar=('SPEC', 'PATTERN'),
                            help=("The inputs to include in the study init. "
                                  "If not provided then all are used"))
        parser.add_argument('--parameter', '-p', nargs=2, action='append',
                            metavar=('NAME', 'VALUE'), default=[],
                            help="Parameters to pass to the study")
        parser.add_argument('--subject_ids', nargs='+', default=None,
                            metavar=['ID'],
                            help=("The subject IDs to include in the analysis."
                                  " If a single value with a '/' in it is "
                                  "provided then it is interpreted as a text "
                                  "file containing a list of IDs"))
        parser.add_argument('--visit_ids', nargs='+', default=None,
                            metavar=['ID'],
                            help=("The visit IDs to include in the analysis"
                                  "If a single value with a '/' in it is "
                                  "provided then it is interpreted as a text "
                                  "file containing a list of IDs"))
        parser.add_argument('--scratch', type=str, default=None,
                            metavar=['PATH'],
                            help=("The scratch directory to use for the "
                                  "workflow and cache"))
        parser.add_argument('--cache', nargs='+', default=(), metavar=['SPEC'],
                            help=("Input filesets to cache locally before "
                                  "running workflows"))
        parser.add_argument('--enforce_inputs', action='store_true',
                            default=False,
                            help=("Whether to enforce inputs for non-optional "
                                  "specs"))
        parser.add_argument('--reprocess', type=bool, default=False,
                            help=("Whether to reprocess the "))
        parser.add_argument('--email', type=str, default=None,
                            help=("The email account to provide to SLURM "
                                  "scheduler"))
        parser.add_argument('--logger', nargs=2, action='append',
                            metavar=['LOGGER', 'LEVEL'],
                            default=[('banana', 'INFO'), ('arcana', 'INFO'),
                                     ('nipype.workflow', 'INFO')],
                            help=("Set levels for various loggers ('arcana', "
                                  "'banana', and 'nipype.workflow' are set to "
                                  "INFO by default)"))
        return parser

    @classmethod
    def run(cls, args):

        set_loggers(args.logger)

        study_class = resolve_class(args.study_class)

        if args.scratch is not None:
            scratch_dir = args.scratch
        else:
            scratch_dir = op.join(op.expanduser('~'), 'banana-scratch')

        # Ensure scratch dir exists
        os.makedirs(scratch_dir, exist_ok=True)

        work_dir = op.join(scratch_dir, 'work')

        if args.repository is None:
            if args.input:
                repo_type = 'basic'
            else:
                repo_type = 'bids'
        else:
            repo_type = args.repository[0]

        if args.repository[0] == 'bids':
            repository = BidsRepo(args.repository_path)
        elif args.repository == 'basic':
            if len(args.repository) != 2:
                raise BananaUsageError(
                    "Unrecognised arguments passed to '--repository' option "
                    "({}) exactly 1 additional argument is required for "
                    "'basic' type repository (DEPTH)".format(args.respository))
            repository = BasicRepo(args.repository_path)
        elif repo_type == 'xnat':
            nargs = len(args.repository)
            if nargs < 2:
                raise BananaUsageError(
                    "Not enough arguments passed to '--repository' option "
                    "({}), at least 1 additional argument is required for "
                    "'xnat' type repository (SERVER)"
                    .format(args.respository))
            elif nargs > 4:
                raise BananaUsageError(
                    "Unrecognised arguments passed to '--repository' option "
                    "({}), at most 3 additional argument are accepted for "
                    "'xnat' type repository (SERVER, USER, PASSWORD)"
                    .format(args.respository))
            repository = XnatRepo(
                project_id=args.repository_path,
                server=args.repository[1],
                user=(args.repository[2] if nargs > 2 else None),
                password=(args.repository[3] if nargs > 3 else None),
                cache_dir=op.join(scratch_dir, 'cache'))
        else:
            raise BananaUsageError(
                "Unrecognised repository type provided as first argument to "
                "'--repository' option ({})".format(args.repository[0]))

        if args.email is not None:
            email = args.email
        else:
            try:
                email = os.environ['EMAIL']
            except KeyError:
                email = None

        proc_args = {'reprocess': args.reprocess}

        if args.processor[0] == 'single':
            processor = SingleProc(work_dir, **proc_args)
        elif args.processor[0] == 'multi':
            if len(args.processor) > 1:
                num_processes = args.processor[1]
            elif len(args.processor) > 2:
                raise BananaUsageError(
                    "Unrecognised arguments passed to '--processor' option "
                    "({}) expected at most 1 additional argument for 'multi' "
                    "type processor (NUM_PROCS)".format(args.processor))
            else:
                num_processes = cpu_count()
            processor = MultiProc(work_dir, num_processes=num_processes,
                                  **proc_args)
        elif args.processor[0] == 'slurm':
            if email is None:
                raise BananaUsageError(
                    "Email needs to be provided either via '--email' argument "
                    "or set in 'EMAIL' environment variable for SLURM "
                    "processor")
            nargs = len(args.processor)
            if nargs > 3:
                raise BananaUsageError(
                    "Unrecognised arguments passed to '--processor' option "
                    "with 'slurm' type ({}), expected at most 2 additional "
                    "arguments [ACCOUNT, PARTITION]".format(args.processor))
            processor = SlurmProc(
                work_dir, account=(args.processor[1] if nargs >= 2 else None),
                partition=(args.processor[2] if nargs >= 3 else None),
                email=email, mail_on=('FAIL',),
                **proc_args)
        else:
            raise BananaUsageError(
                "Unrecognised processor type provided as first argument to "
                "'--processor' option ({})".format(args.processor[0]))

        if args.environment == 'static':
            environment = StaticEnv()
        else:
            environment = ModulesEnv()

        # Load subject_ids from file if single value is provided with
        # a '/' in the string
        if (args.subject_ids is not None and len(args.subject_ids) and
                '/' in args.subject_ids[0]):
            with open(args.subject_ids[0]) as f:
                subject_ids = f.read().split()
        else:
            subject_ids = args.subject_ids

        # Load visit_ids from file if single value is provided with
        # a '/' in the string
        if (args.visit_ids is not None and len(args.visit_ids) and
                '/' in args.visit_ids[0]):
            with open(args.visit_ids[0]) as f:
                visit_ids = f.read().split()
        else:
            visit_ids = args.visit_ids

        parameters = {}
        for name, value in args.parameter:
            parameters[name] = parse_value(value)

        study = study_class(
            name=args.study_name,
            repository=repository,
            processor=processor,
            environment=environment,
            inputs=dict(args.input),
            parameters=parameters,
            subject_ids=subject_ids,
            visit_ids=visit_ids,
            enforce_inputs=args.enforce_inputs)

        for spec_name in args.cache:
            spec = study.bound_spec(spec_name)
            if not isinstance(spec, InputFilesets):
                raise BananaUsageError(
                    "Cannot cache non-input fileset '{}'".format(spec_name))
            spec.cache()

        # Generate data
        study.data(args.derivatives)

        logger.info("Generated derivatives for '{}'".format(args.derivatives))


class TestDataCmd():

    desc = "Generate derivatives from a study"

    @classmethod
    def parser(cls):
        parser = ArgumentParser(
            prog='banana test_data',
            description=("Generates reference data for a pipeline tester "
                         "unittests given a study class and set of "
                         "parameters"))
        parser.add_argument('study_class',
                            help=("The path to the study class to test, e.g. "
                                  "banana.study.MriStudy"))
        parser.add_argument('in_repo', help=("The path to repository that "
                                             "houses the input data"))
        parser.add_argument('out_repo',
                            help=("If the 'xnat_server' argument is provided "
                                  "then out is interpreted as the project ID "
                                  "to use the XNAT server (the project must "
                                  "exist already). Otherwise it is interpreted"
                                  " as the path to a basic repository"))
        parser.add_argument('--in_server', default=None,
                            help="The server to download the input data from")
        parser.add_argument('--out_server', default=None,
                            help="The server to upload the reference data to")
        parser.add_argument('--work_dir', default=None,
                            help="The work directory")
        parser.add_argument('--parameter', '-p', metavar=('NAME', 'VALUE'),
                            nargs=2, action='append', default=[],
                            help=("Parameters to set when initialising the "
                                  "study"))
        parser.add_argument('--skip', '-s', nargs='+', default=[],
                            help=("Spec names to skip in the generation "
                                  "process"))
        parser.add_argument('--skip_base', action='append', default=[],
                            help=("Base classes of which to skip data specs "
                                  "from"))
        parser.add_argument('--reprocess', action='store_true', default=False,
                            help=("Whether to reprocess previously generated "
                                  "datasets in the output repository"))
        parser.add_argument('--repo_depth', type=int, default=0,
                            help="The depth of the input repository")
        parser.add_argument('--modules_env', action='store_true',
                            default=False,
                            help="Whether to use a Modules Envionment or not")
        parser.add_argument('--dont_clean_work_dir', action='store_true',
                            default=False,
                            help=("Whether to clean the Nipype work dir "
                                  "between runs"))
        parser.add_argument('--loggers', nargs='+',
                            default=('nipype.workflow', 'arcana', 'banana'),
                            help="Loggers to set handlers to stdout for")
        return parser

    @classmethod
    def run(cls, args):

        # Get Study class
        study_class = resolve_class(args.study_class)

        include_bases = [resolve_class(c) for c in args.skip_base]

        # Convert parameters to dictionary
        parameters_dct = {}
        for name, value in args.parameter:
            try:
                value = int(value)
            except ValueError:
                try:
                    value = float(value)
                except ValueError:
                    pass
            parameters_dct[name] = value
        parameters = parameters_dct

        PipelineTester.generate_test_data(
            study_class=study_class, in_repo=args.in_repo,
            out_repo=args.out_repo, in_server=args.in_server,
            out_server=args.out_server, work_dir=args.work_dir,
            parameters=parameters, skip=args.skip, include_bases=include_bases,
            reprocess=args.reprocess, repo_depth=args.repo_depth,
            modules_env=args.modules_env,
            clean_work_dir=(not args.dont_clean_work_dir))


class HelpCmd():

    desc = "Show help for a particular command"

    @classmethod
    def parser(cls):
        parser = ArgumentParser(prog='banana help',
                                description=cls.desc)
        parser.add_argument('command',
                            help="The sub-command to show the help info for")
        return parser

    @classmethod
    def run(cls, argv):
        args = cls.parser().parse_args(argv)
        MainCmd.commands[args.command].parser().print_help()


class MenuCmd():

    desc = ("Display the data and parameter specifications for a given study "
            "class")

    @classmethod
    def parser(cls):
        parser = ArgumentParser(prog='banana menu',
                                description=cls.desc)
        parser.add_argument('study_class',
                            help=("Name of the class to display menu for"))
        return parser

    @classmethod
    def run(cls, args):
        # Get Study class
        study_class = resolve_class(args.study_class)
        print(study_class.static_menu())


class MainCmd():

    commands = {
        'derive': DeriveCmd,
        'menu': MenuCmd,
        'test_data': TestDataCmd,
        'help': HelpCmd}

    @classmethod
    def parser(cls):
        usage = "banana <command> [<args>]\n\nAvailable commands:"
        for name, cmd_cls in cls.commands.items():
            usage += '\n\t{}\t\t{}'.format(name, cmd_cls.desc)
        parser = ArgumentParser(
            description="Base banana command",
            usage=usage)
        parser.add_argument('command', help="The sub-command to run")
        return parser

    @classmethod
    def run(cls, argv=None):
        if argv is None:
            argv = sys.argv[1:]
        parser = cls.parser()
        args = parser.parse_args(argv[:1])
        try:
            cmd_cls = cls.commands[args.command]
        except KeyError:
            print("Unrecognised command '{}'".format(args.command))
            parser.print_help()
            exit(1)
        cmd_args = cmd_cls.parser().parse_args(argv[1:])
        cmd_cls.run(cmd_args)


if __name__ == '__main__':
    MainCmd.run()



from future.utils import PY3
import os
import os.path as op
import platform
import re
import xml.etree.ElementTree
from arcana.environment.requirement import (
    Version, CliRequirement, MatlabPackageRequirement,
    PythonPackageRequirement,
    matlab_req)  # @UnusedImport
from arcana.utils import run_matlab_cmd
from arcana.exceptions import (
    ArcanaRequirementNotFoundError, ArcanaVersionNotDetectableError)

# Command line requirements


class FSLRequirement(CliRequirement):

    def detect_version_str(self):
        """
        As FSL doesn't have a simple way of printing the version, we need to
        read the version from the 'fslversion' text file in the FSLDIR.
        """
        try:
            fsl_dir = os.environ['FSLDIR']
        except KeyError:
            raise ArcanaRequirementNotFoundError(
                "Could not find FSL, 'FSLDIR' environment variable is not set")
        with open(op.join(fsl_dir, 'etc', 'fslversion'), 'r') as f:
            contents = f.read()
        return contents.strip()


class C3dRequirement(CliRequirement):

    def detect_version_str(self):
        """
        C3D doesn't print out the version so we need to interrogate the
        install directory to extract it
        """
        c3d_bin_path = op.dirname(self.locate_command())
        if platform.system() == 'Linux':
            libname = os.listdir(op.join(c3d_bin_path, '..', 'lib'))[0]
            version_str = libname.split('-')[-1]
        elif platform.system() == 'Darwin':
            info_list_path = op.join(c3d_bin_path, '..', 'Info.plist')
            info_etree = xml.etree.ElementTree.parse(info_list_path)
            elem_bodies = [e.text for e in info_etree.iter()]
            version_str = elem_bodies[
                elem_bodies.index('CFBundleShortVersionString') + 1]
        else:
            raise ArcanaVersionNotDetectableError(
                "Can't detect c3d version on Windows")
        return version_str


class FreesurferRequirement(CliRequirement):

    def detect_version_str(self):
        """
        The version that recon-all spits out doesn't match that of Freesurfer
        itself so we take it from the build-stamp.txt file instead
        """
        try:
            fs_dir = os.environ['FREESURFER_HOME']
        except KeyError:
            raise ArcanaRequirementNotFoundError(
                "Could not find Freesurfer installation as 'FREESURFER_HOME' "
                "environment variable is not set")
        with open(op.join(fs_dir, 'build-stamp.txt'), 'r') as f:
            contents = f.read()
        return re.match(r'freesurfer-.*-v(.*)', contents).group(1)


class MrtrixRequirement(CliRequirement):

    def detect_version_str(self):
        version_str = super().detect_version_str()
        return re.match(r'== mrinfo (.*) ==', version_str).group(1)


class AfniRequirement(CliRequirement):

    def detect_version_str(self):
        version_str = super().detect_version_str()
        return re.match(r'.*AFNI_([\d\.]+)', version_str).group(1)


class FixVersion(Version):

    regex = re.compile(r'(\d+)\.(\d)?(\d)?(\d)?')

    def __str__(self):
        return '{}.{}'.format(self._seq[0], ''.join(self._seq[1:]))


class StirRequirement(CliRequirement):

    def detect_version_str(self):
        raise ArcanaVersionNotDetectableError(
            "Can't automatically detect version of STIR as it isn't saved in "
            "the build process")


mrtrix_req = MrtrixRequirement('mrtrix', test_cmd='mrinfo')
ants_req = CliRequirement('ants', test_cmd='antsRegistration')
dcm2niix_req = CliRequirement('dcm2niix', test_cmd='dcm2niix')
freesurfer_req = FreesurferRequirement('freesurfer', test_cmd='recon-all')
fix_req = CliRequirement('fix', test_cmd='fix', version_cls=FixVersion)
afni_req = AfniRequirement('afni', test_cmd='afni')
fsl_req = FSLRequirement('fsl', test_cmd='fslinfo')
c3d_req = C3dRequirement('c3d', test_cmd='c3d')
stir_req = StirRequirement('stir', test_cmd='SSRB')

# Matlab package requirements


class SpmRequirement(MatlabPackageRequirement):

    def parse_help_text(self, help_text):
        """
        Detect which SPM version we are using from the copyright year
        """
        match = re.search(
            r'Copyright \(C\) [\d\-\, ]*(?<!\d)(\d+) Wellcome Trust Centre',
            help_text)
        if match is None:
            raise ArcanaVersionNotDetectableError(
                "Could not parse year of copyright from spm_authors in order "
                "to determine the version of {}".format(self))
        copyright_year = match.group(1)
        if copyright_year == '2010':
            version = 8
        elif copyright_year == '2012':
            version = 12
        else:
            raise ArcanaVersionNotDetectableError(
                "Do not know the version of SPM corresponding to the year of "
                "copyright of {}".format(copyright_year))
        return version


class StiRequirement(MatlabPackageRequirement):

    def detect_version_str(self):
        """
        Detect which version by scanning the README for the latest release
        """
        cmd_path = run_matlab_cmd("which('{}')".format(self.test_func))
        pkg_root = op.join(op.dirname(cmd_path), '..')
        readmes = [f for f in os.listdir(pkg_root) if 'readme' in f.lower()]
        if not readmes:
            raise ArcanaVersionNotDetectableError(
                "Did not find a README in STI package root ({})"
                .format(pkg_root))
        elif len(readmes) > 1:
            raise ArcanaVersionNotDetectableError(
                "Found multiple READMEs in STI package root ({})"
                .format(pkg_root))
        readme_path = op.join(pkg_root, readmes[0])
        with open(readme_path, 'rb') as f:
            contents = f.read()
        # Cut out citations text as there can be non-decodable characters in
        # there
        contents = contents.split(b'TERMS OF USE')[0]
        if PY3:
            contents = contents.decode('utf-8')
        # Get dummy version so we can use its 'regex' property
        dummy_version_obj = self.version_cls(self, 1)
        versions = dummy_version_obj.regex.findall(contents)
        latest_version = sorted(versions)[-1]
        return latest_version


spm_req = SpmRequirement('spm', test_func='spm_authors')
sti_req = StiRequirement('sti', test_func='V_SHARP')
# noddi_req = MatlabPackageRequirement('noddi')


# Python package requirements

sklearn_req = PythonPackageRequirement('sklearn')
pydicom_req = PythonPackageRequirement('pydicom')
scipy_req = PythonPackageRequirement('scipy')



import logging
import os.path as op
from nipype.interfaces.base import (
    TraitedSpec, traits, BaseInterface, BaseInterfaceInputSpec, Directory,
    File)
logger = logging.getLogger('banana')


class AppendPathInputSpec(BaseInterfaceInputSpec):
    base_path = Directory(exists=True, mandatory=True)
    sub_paths = traits.List(
        traits.Str, desc=("The sub-paths to append to the "))


class AppendPathOutputSpec(TraitedSpec):
    out_path = traits.Either(
        File, Directory, desc="The first echo of the combined images")


class AppendPath(BaseInterface):
    """
    Appends a sub-path to an existing directory
    """
    input_spec = AppendPathInputSpec
    output_spec = AppendPathOutputSpec

    def _run_interface(self, runtime):
        return runtime

    def _list_outputs(self):  # @UnusedVariable
        outputs = self._outputs().get()
        outputs['out_path'] = op.join(self.inputs.base_path,
                                      *self.inputs.sub_paths)
        return outputs



import os
import json
import os.path as op
import stat
import logging
from bids.layout import BIDSLayout
from arcana.exceptions import (
    ArcanaInputMissingMatchError, ArcanaUsageError)
from banana.exceptions import BananaUsageError
from arcana.data.input import InputFilesets
from arcana.data.item import Fileset
from arcana.data.file_format import FileFormat
from arcana.utils import split_extension
from arcana.repository import BasicRepo
from banana.file_format import (
    nifti_gz_format, nifti_gz_x_format, fsl_bvecs_format, fsl_bvals_format,
    tsv_format)


logger = logging.getLogger('arcana')

BIDS_FORMATS = (nifti_gz_x_format, nifti_gz_format, fsl_bvecs_format,
                fsl_bvals_format, tsv_format)


def detect_format(fileset):
    for frmt in BIDS_FORMATS:
        if frmt.matches(fileset):
            return frmt
    raise BananaUsageError(
        "No matching BIDS format found for {}".format(fileset))


class BidsRepo(BasicRepo):
    """
    A repository class for BIDS datasets

    Parameters
    ----------
    root_dir : str
        The path to the root of the BidsRepo
    """

    type = 'bids'

    def __init__(self, root_dir, **kwargs):
        BasicRepo.__init__(self, root_dir, depth=2, **kwargs)
        self._layout = BIDSLayout(root_dir)

    @property
    def root_dir(self):
        return self._root_dir

    @property
    def derivatives_dir(self):
        return op.join(self.root_dir, 'derivatives')

    @property
    def metadata_dir(self):
        """
        A temporary dir where we write out combined JSON side cars to include
        in extended nifti filesets
        """
        return op.join(self.derivatives_dir, '__metadata__')

    @property
    def layout(self):
        return self._layout

    def __repr__(self):
        return "BidsRepo(root_dir='{}')".format(self.root_dir)

    def __hash__(self):
        return super().__hash__()

    def find_data(self, subject_ids=None, visit_ids=None):
        """
        Return subject and session information for a project in the local
        repository

        Parameters
        ----------
        subject_ids : list(str)
            List of subject IDs with which to filter the tree with. If None all
            are returned
        visit_ids : list(str)
            List of visit IDs with which to filter the tree with. If None all
            are returned

        Returns
        -------
        project : arcana.repository.Tree
            A hierarchical tree of subject, session and fileset information for
            the repository
        """
        filesets = []
        all_subjects = self.layout.get_subjects()
        all_visits = self.layout.get_sessions()
        for item in self.layout.get(return_type='object'):
            if item.path.startswith(self.derivatives_dir):
                # We handle derivatives using the BasicRepo base
                # class methods
                continue
            if not hasattr(item, 'entities') or not item.entities.get('type',
                                                                      False):
                logger.warning("Skipping unrecognised file '{}' in BIDS tree"
                               .format(op.join(item.dirname, item.filename)))
                continue  # Ignore hidden file
            try:
                subject_ids = [item.entities['subject']]
            except KeyError:
                # If item exists in top-levels of in the directory structure
                # it is inferred to exist for all subjects in the tree
                subject_ids = all_subjects
            try:
                visit_ids = [item.entities['session']]
            except KeyError:
                # If item exists in top-levels of in the directory structure
                # it is inferred to exist for all visits in the tree
                visit_ids = all_visits
            for subject_id in subject_ids:
                for visit_id in visit_ids:
                    aux_files = {}
                    metadata = self.layout.get_metadata(item.path)
                    if metadata and not item.path.endswith('.json'):
                        # Write out the combined JSON side cars to a temporary
                        # file to include in extended NIfTI filesets
                        metadata_path = op.join(
                            self.metadata_dir,
                            'sub-{}'.format(subject_id),
                            'ses-{}'.format(visit_id),
                            item.filename + '.json')
                        os.makedirs(op.dirname(metadata_path), exist_ok=True)
                        if not op.exists(metadata_path):
                            with open(metadata_path, 'w') as f:
                                json.dump(metadata, f)
                        aux_files['json'] = metadata_path
                    fileset = BidsFileset(
                        path=op.join(item.dirname, item.filename),
                        type=item.entities['type'],
                        subject_id=subject_id, visit_id=visit_id,
                        repository=self,
                        modality=item.entities.get('modality', None),
                        task=item.entities.get('task', None),
                        aux_files=aux_files)
                    filesets.append(fileset)
        # Get derived filesets, fields and records using the same method using
        # the method in the BasicRepo base class
        derived_filesets, fields, records = super().find_data(
            subject_ids=subject_ids, visit_ids=visit_ids)
        filesets.extend(derived_filesets)
        return filesets, fields, records

    def fileset_path(self, fileset, fname=None):
        if not fileset.derived:
            raise ArcanaUsageError(
                "Can only get automatically get path to derived filesets not "
                "{}".format(fileset))
        if fname is None:
            fname = fileset.fname
        if fileset.subject_id is not None:
            subject_id = fileset.subject_id
        else:
            subject_id = self.SUMMARY_NAME
        if fileset.visit_id is not None:
            visit_id = fileset.visit_id
        else:
            visit_id = self.SUMMARY_NAME
        sess_dir = op.join(self.root_dir, 'derivatives', fileset.from_study,
                           'sub-{}'.format(subject_id),
                           'sess-{}'.format(visit_id))
        # Make session dir if required
        if not op.exists(sess_dir):
            os.makedirs(sess_dir, stat.S_IRWXU | stat.S_IRWXG)
        return op.join(sess_dir, fname)

    def _extract_ids_from_path(self, path_parts, *args, **kwargs):  # @UnusedVariable @IgnorePep8
        if len(path_parts) != 4 or path_parts[0] != 'derivatives':
            return None
        from_study, subj, sess = path_parts[1:]
        subj_id = subj[len('sub-'):]
        visit_id = sess[len('sess-'):]
        return subj_id, visit_id, from_study


class BaseBidsFileset(object):

    derived = False

    def __init__(self, type, modality, task):  # @ReservedAssignment
        self._modality = modality
        self._type = type
        self._task = task

    def __eq__(self, other):
        return (self.type == other.type and
                self.task == other.task and
                self.modality == other.modality)

    def __hash__(self):
        return (hash(self.type) ^
                hash(self.task) ^
                hash(self.modality))

    def initkwargs(self):
        dct = {}
        dct['type'] = self.type
        dct['task'] = self.task
        dct['modality'] = self.modality
        return dct

    @property
    def modality(self):
        return self._modality

    @property
    def task(self):
        return self._task

    @property
    def type(self):
        return self._type


class BidsFileset(Fileset, BaseBidsFileset):
    """
    A representation of a fileset within the repository.

    Parameters
    ----------
    name : str
        The name of the fileset
    path : str | None
        The path to the fileset on the local system
    subject_id : int | str | None
        The id of the subject which the fileset belongs to
    visit_id : int | str | None
        The id of the visit which the fileset belongs to
    repository : Repository
        The repository which the fileset is stored
    modality : str
        The BIDS modality
    task : str
        The BIDS task
    checksums : dict[str, str]
        A checksums of all files within the fileset in a dictionary sorted by
        relative file paths
    aux_files : dict[str, str]
        A dictionary containing a mapping from a side car name to path of the
        file
    """

    def __init__(self, path, type, subject_id, visit_id, repository,  # @ReservedAssignment @IgnorePep8
                 modality=None, task=None, checksums=None, aux_files=None):
        Fileset.__init__(
            self,
            name=op.basename(path),
            frequency='per_session',
            path=path,
            subject_id=subject_id,
            visit_id=visit_id,
            repository=repository,
            checksums=checksums,
            aux_files=aux_files)
        BaseBidsFileset.__init__(self, type, modality, task)
        self.detect_format(BIDS_FORMATS)

    def __repr__(self):
        return ("{}(type={}, task={}, modality={}, format={}, subj={}, vis={})"
                .format(self.__class__.__name__, self.type, self.task,
                        self.modality, self.format.name, self.subject_id,
                        self.visit_id))


class BidsInput(InputFilesets, BaseBidsFileset):
    """
    A match object for matching filesets from their BIDS attributes and file
    format. If any of the provided attributes are None, then that attribute
    is omitted from the match

    Parameters
    ----------
    spec_name : str
        Name of the spec to match
    type : str
        Type of the fileset
    valid_formats : FileFormat | list(FileFormat)
        The file format of the fileset to match, or a list of valid formats
    task : str
        The task the fileset belongs to
    modality : str
        Modality of the filesets
    """

    def __init__(self, spec_name, type, format=None, task=None, modality=None,  # @ReservedAssignment @IgnorePep8
                 **kwargs):
        InputFilesets.__init__(
            self, spec_name, pattern=None, format=format,
            frequency='per_session', **kwargs)  # @ReservedAssignment @IgnorePep8
        BaseBidsFileset.__init__(self, type, modality, task)

    def _filtered_matches(self, node):
        matches = [
            f for f in node.filesets
            if (isinstance(f, BidsFileset) and
                self.type == f.type and
                (self.modality is None or self.modality == f.modality) and
                (self.task is None or self.task == f.task) and
                (self.format is None or self.format == f.format))]
        if not matches:
            raise ArcanaInputMissingMatchError(
                "No BIDS filesets for {} match {} found:\n{}"
                .format(node, self, '\n'.join(str(f) for f in node.filesets)))
        return matches

    def __repr__(self):
        return ("{}(spec_name='{}', type={}, format={}, modality={}, task={})"
                .format(
                    self.__class__.__name__,
                    self.spec_name,
                    self.type,
                    self._format.name if self._format is not None else None,
                    self.modality, self.task))

    def __eq__(self, other):
        return (InputFilesets.__eq__(self, other) and
                BaseBidsFileset.__eq__(self, other))

    def __hash__(self):
        return (InputFilesets.__hash__(self) ^
                BaseBidsFileset.__hash__(self))

    def initkwargs(self):
        dct = InputFilesets.initkwargs(self)
        dct.update(BaseBidsFileset.initkwargs(self))
        return dct

    def _check_args(self):
        pass  # Disable check for either pattern or ID in base class

    @BaseBidsFileset.task.setter
    def task(self, task):
        self._task = task


class BidsAssocInput(InputFilesets):
    """
    A match object for matching BIDS filesets that are associated with
    another BIDS filesets (e.g. field-maps, bvecs, bvals)

    Parameters
    ----------
    name : str
        Name of the associated fileset
    primary : BidsInput
        A selector to select the primary fileset which the associated fileset
        is associated with
    association : str
        The name of the association between the fileset to match and the
        primary fileset, can be one of 'bvec', 'bval', 'phase1', 'phase2',
        'phasediff', 'epi' or 'fieldmap'
    type : int
        If there are more than one field-maps associated with the primary
        fileset, which one to return
    """

    VALID_ASSOCIATIONS = ('grads', 'phase', 'phasediff', 'epi', 'fieldmap')

    def __init__(self, spec_name, primary, association, type=None, format=None,   # @ReservedAssignment @IgnorePep8
                 **kwargs):
        InputFilesets.__init__(self, spec_name, format,
                                 frequency='per_session', **kwargs)
        self._primary = primary
        if association not in self.VALID_ASSOCIATIONS:
            raise BananaUsageError(
                "Invalid association '{}' passed to BidsAssocInput, "
                "can be one of '{}'".format(
                    association, "', '".join(self.VALID_ASSOCIATIONS)))
        self._association = association
        self._type = type

    def __eq__(self, other):
        return (InputFilesets.__eq__(self, other) and
                self.primary == other.primary and
                self.format == other.format and
                self.association == other.association and
                self._type == other._type)

    def __hash__(self):
        return (InputFilesets.__hash__(self) ^
                hash(self.primary) ^
                hash(self.format) ^
                hash(self.association) ^
                hash(self._type))

    def initkwargs(self):
        dct = InputFilesets.initkwargs(self)
        dct['primary'] = self.primary
        dct['format'] = self.primary
        dct['association'] = self.association
        dct['type'] = self._type
        return dct

    def __repr__(self):
        return ("{}(spec_name={}, primary={}, format={}, association={}, "
                "type={})".format(
                    type(self).__name__,
                    self.spec_name, self.primary,
                    self._format.name if self._format is not None else None,
                    self.association, self.type))

    def bind(self, study, spec_name=None, **kwargs):
        # We need to access a bound primary selector when matching the
        # associated selector so we set the bound version temporarily to
        # self._primary before winding it back after we have done the bind
        unbound_primary = self._primary
        self._primary = self._primary.bind(study, **kwargs)
        bound = super().bind(study, spec_name=spec_name, **kwargs)
        self._primary = unbound_primary
        return bound

    @property
    def primary(self):
        return self._primary

    @property
    def association(self):
        return self._association

    @property
    def type(self):
        return self._type if self._type is not None else self.association

    @property
    def task(self):
        return self.primary.task

    @task.setter
    def task(self, task):
        self.primary.task = task

    def match_node(self, node):
        primary_match = self.primary.match_node(node)
        layout = self.primary.repository.layout
        if self.association == 'grads':
            if self.type == 'bvec':
                path = layout.get_bvec(primary_match.path)
            elif self.type == 'bval':
                path = layout.get_bval(primary_match.path)
            else:
                raise ArcanaUsageError(
                    "'{}' is not a valid type for '{}' associations"
                    .format(self.type, self.association))
        else:
            fieldmaps = layout.get_fieldmap(primary_match.path,
                                            return_list=True)
            try:
                fieldmap = next(f for f in fieldmaps
                                     if f['type'] == self.association)
            except StopIteration:
                raise ArcanaInputMissingMatchError(
                    "No \"{}\" field-maps associated with {} (found {})"
                    .format(self.association, primary_match,
                            ', '.join(f['type'] for f in fieldmaps)))
            try:
                path = fieldmap[self.type]
            except KeyError:
                raise ArcanaUsageError(
                    "'{}' is not a valid type for '{}' associations"
                    .format(self.type, self.association))
        return Fileset.from_path(path, format=self._format,
                                 repository=self.primary.repository,
                                 subject_id=node.subject_id,
                                 visit_id=node.visit_id)



from arcana.exceptions import ArcanaError


class BananaError(ArcanaError):
    pass


class BananaUsageError(BananaError):
    pass


class BananaMissingHeaderValue(BananaError):
    pass


class BananaRuntimeError(BananaError):
    pass


class BananaTestSetupError(BananaError):
    pass



from copy import copy
import itertools
from nipype.interfaces.freesurfer.preprocess import ReconAll
# from arcana.utils.interfaces import DummyReconAll as ReconAll
from nipype.interfaces import fsl, ants, mrtrix3
from arcana.utils.interfaces import Merge
from arcana.data import FilesetSpec, InputFilesetSpec
from arcana.utils.interfaces import JoinPath
from arcana.study.base import StudyMetaClass
from arcana.study import ParamSpec, SwitchSpec
from arcana.utils.interfaces import CopyToDir
from banana.requirement import freesurfer_req, ants_req, fsl_req, mrtrix_req
from banana.citation import freesurfer_cites, fsl_cite
from banana.interfaces.freesurfer import AparcStats
from banana.interfaces.utility import AppendPath
from banana.file_format import (
    nifti_gz_format, nifti_format, zip_format, STD_IMAGE_FORMATS,
    directory_format, text_format, mrtrix_image_format)
from banana.reference import LocalReferenceData
from .t2 import T2Study


class T1Study(T2Study, metaclass=StudyMetaClass):

    add_data_specs = [
        FilesetSpec('fs_recon_all', zip_format, 'freesurfer_pipeline'),
        InputFilesetSpec(
            't2_coreg', STD_IMAGE_FORMATS, optional=True,
            desc=("A coregistered T2 image to use in freesurfer to help "
                  "distinguish the peel surface")),
        # Templates
        InputFilesetSpec('suit_mask', STD_IMAGE_FORMATS,
                         frequency='per_study',
                         default=LocalReferenceData('SUIT', nifti_format)),
        FilesetSpec('five_tissue_type', mrtrix_image_format,
                    'gen_5tt_pipeline',
                    desc=("A segmentation image taken from freesurfer output "
                          "and simplified into 5 tissue types. Used in ACT "
                          "streamlines tractography"))] + [
        FilesetSpec(
            'aparc_stats_{}_{}_table'.format(h, m),
            text_format, 'aparc_stats_table_pipeline',
            frequency='per_visit',
            pipeline_args={'hemisphere': h, 'measure': m},
            desc=("Table of {} of {} per parcellated segment"
                  .format(m, h.upper())))
        for h, m in itertools.product(
            ('lh', 'rh'),
            ('volume', 'thickness', 'thicknessstd', 'meancurv', 'gauscurv',
             'foldind', 'curvind'))]

    add_param_specs = [
        # MriStudy.param_spec('bet_method').with_new_choices(default='opti_bet'),
        SwitchSpec('bet_robust', False),
        SwitchSpec('bet_reduce_bias', True),
        SwitchSpec('aparc_atlas', 'desikan-killiany',
                   choices=('desikan-killiany', 'destrieux', 'DKT')),
        ParamSpec('bet_f_threshold', 0.1),
        ParamSpec('bet_g_threshold', 0.0)]

    def freesurfer_pipeline(self, **name_maps):
        """
        Segments grey matter, white matter and CSF from T1 images using
        SPM "NewSegment" function.

        NB: Default values come from the W2MHS toolbox
        """
        pipeline = self.new_pipeline(
            name='segmentation',
            name_maps=name_maps,
            desc="Segment white/grey matter and csf",
            citations=copy(freesurfer_cites))

        # FS ReconAll node
        recon_all = pipeline.add(
            'recon_all',
            interface=ReconAll(
                directive='all',
                openmp=self.processor.num_processes),
            inputs={
                'T1_files': ('mag_preproc', nifti_gz_format)},
            requirements=[freesurfer_req.v('5.3')],
            wall_time=2000)

        if self.provided('t2_coreg'):
            pipeline.connect_input('t2_coreg', recon_all, 'T2_file',
                                   nifti_gz_format)
            recon_all.inputs.use_T2 = True

        # Wrapper around os.path.join
        pipeline.add(
            'join',
            JoinPath(),
            inputs={
                'dirname': (recon_all, 'subjects_dir'),
                'filename': (recon_all, 'subject_id')},
            outputs={
                'fs_recon_all': ('path', directory_format)})

        return pipeline

    def segmentation_pipeline(self, **name_maps):
        pipeline = super(T1Study, self).segmentation_pipeline(img_type=1,
                                                              **name_maps)
        return pipeline

    def gen_5tt_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='gen5tt',
            name_maps=name_maps,
            desc=("Generate 5-tissue-type image used for Anatomically-"
                  "Constrained Tractography (ACT)"))

        aseg_path = pipeline.add(
            'aseg_path',
            AppendPath(
                sub_paths=['mri', 'aseg.mgz']),
            inputs={
                'base_path': ('fs_recon_all', directory_format)})

        pipeline.add(
            'gen5tt',
            mrtrix3.Generate5tt(
                algorithm='freesurfer',
                out_file='5tt.mif'),
            inputs={
                'in_file': (aseg_path, 'out_path')},
            outputs={
                'five_tissue_type': ('out_file', mrtrix_image_format)},
            requirements=[mrtrix_req.v('3.0rc3'),
                          freesurfer_req.v('6.0')])

        return pipeline

    def aparc_stats_table_pipeline(self, measure, hemisphere, **name_maps):

        pipeline = self.new_pipeline(
            name='aparc_stats_{}_{}'.format(hemisphere, measure),
            name_maps=name_maps,
            desc=("Extract statistics from freesurfer outputs"))

        copy_to_dir = pipeline.add(
            'copy_to_subjects_dir',
            CopyToDir(),
            inputs={
                'in_files': ('fs_recon_all', directory_format),
                'file_names': (self.SUBJECT_ID, int)},
            joinsource=self.SUBJECT_ID,
            joinfield=['in_files', 'file_names'])

        if self.branch('aparc_atlas', 'desikan-killiany'):
            parc = 'aparc'
        elif self.branch('aparc_atlas', 'destrieux'):
            parc = 'aparc.a2009s'
        elif self.branch('aparc_atlas', 'DKT'):
            parc = 'aparc.DKTatlas40'
        else:
            self.unhandled_branch('aparc_atlas')

        pipeline.add(
            'aparc_stats',
            AparcStats(
                measure=measure,
                hemisphere=hemisphere,
                parc=parc),
            inputs={
                'subjects_dir': (copy_to_dir, 'out_dir'),
                'subjects': (copy_to_dir, 'file_names')},
            outputs={
                'aparc_stats_{}_{}_table'
                .format(hemisphere, measure): ('tablefile', text_format)},
            requirements=[freesurfer_req.v('5.3')])

        return pipeline

    def bet_T1(self, **name_maps):

        pipeline = self.new_pipeline(
            name='BET_T1',
            name_maps=name_maps,
            desc=("Brain extraction pipeline using FSL's BET"),
            citations=[fsl_cite])

        bias = pipeline.add(
            'n4_bias_correction',
            ants.N4BiasFieldCorrection(),
            inputs={
                'input_image': ('t1', nifti_gz_format)},
            requirements=[ants_req.v('1.9')],
            wall_time=60, mem_gb=12)

        pipeline.add(
            'bet',
            fsl.BET(
                frac=0.15,
                reduce_bias=True,
                output_type='NIFTI_GZ'),
            inputs={
                'in_file': (bias, 'output_image')},
            outputs={
                'betted_T1': ('out_file', nifti_gz_format),
                'betted_T1_mask': ('mask_file', nifti_gz_format)},
            requirements=[fsl_req.v('5.0.8')], mem_gb=8,
            wall_time=45)

        return pipeline

    def cet_T1(self, **name_maps):
        pipeline = self.new_pipeline(
            name='CET_T1',
            name_maps=name_maps,
            desc=("Construct cerebellum mask using SUIT template"),
            citations=[fsl_cite])

        # FIXME: Should convert to inputs
        nl = self._lookup_nl_tfm_inv_name('MNI')
        linear = self._lookup_l_tfm_to_name('MNI')

        # Initially use MNI space to warp SUIT into T1 and threshold to mask
        merge_trans = pipeline.add(
            'merge_transforms',
            Merge(2),
            inputs={
                'in2': (nl, nifti_gz_format),
                'in1': (linear, nifti_gz_format)})

        apply_trans = pipeline.add(
            'ApplyTransform',
            ants.resampling.ApplyTransforms(
                interpolation='NearestNeighbor',
                input_image_type=3,
                invert_transform_flags=[True, False]),
            inputs={
                'reference_image': ('betted_T1', nifti_gz_format),
                'input_image': ('suit_mask', nifti_gz_format),
                'transforms': (merge_trans, 'out')},
            requirements=[ants_req.v('1.9')], mem_gb=16,
            wall_time=120)

        pipeline.add(
            'maths2',
            fsl.utils.ImageMaths(
                suffix='_optiBET_cerebellum',
                op_string='-mas'),
            inputs={
                'in_file': ('betted_T1', nifti_gz_format),
                'in_file2': (apply_trans, 'output_image')},
            outputs={
                'cetted_T1': ('out_file', nifti_gz_format),
                'cetted_T1_mask': ('output_image', nifti_gz_format)},
            requirements=[fsl_req.v('5.0.8')], mem_gb=16,
            wall_time=5)

        return pipeline


class PreclinicalT1(T1Study, metaclass=StudyMetaClass):

    pass



import os.path
from nipype.interfaces.base import (
    traits, InputMultiPath, File, TraitedSpec, isdefined)
from nipype.interfaces.mrtrix3.reconst import (
    MRTrix3Base, MRTrix3BaseInputSpec)
from arcana.utils import split_extension


class MRResizeInputSpec(MRTrix3BaseInputSpec):
    in_file = File(exists=True, argstr='%s', mandatory=True, position=-2,
                   desc='input files')
    out_file = File(genfile=True, argstr='%s', desc=(""), position=-1)

    size = traits.List(
        traits.Int(),
        mandatory=False, argstr='-size %s',
        desc=("define the new image size for the output image. "
              "This should be specified as a comma-separated "
              "list."))

    voxel = traits.List(
        traits.Float(),
        mandatory=False, argstr='-voxel %s',
        desc=("define the new voxel size for the output image."
              "This can be specified either as a single value "
              "to be used for all dimensions, or as a "
              "comma-separated list of the size for each voxel"
              "dimension."))

    scale = traits.Float(
        mandatory=False, argstr='-scale %s',
        desc=("scale the image resolution by the supplied factor."
              " This can be specified either as a single value to"
              " be used for all dimensions, or as a "
              "comma-separated list of scale factors for each "
              "dimension."))

    interp = traits.Enum(
        'nearest', 'linear', 'cubic', 'sinc', default='cubic', mandatory=False,
        argstr='-interp %s',
        desc=("set the interpolation method to use when resizing"
              " (choices: nearest, linear, cubic, sinc. Default:"
              " cubic)."))


class MRResizeOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc=("The output file"))


class MRResize(MRTrix3Base):
    """
    Note that if the image is 4D, then only the first 3 dimensions can be
     resized.

    Also note that if the image is down-sampled, the appropriate smoothing is
    automatically applied using Gaussian smoothing.
    """

    _cmd = "mrresize"
    input_spec = MRResizeInputSpec
    output_spec = MRResizeOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = self._gen_outfilename()
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            gen_name = self._gen_outfilename()
        else:
            assert False
        return gen_name

    def _gen_outfilename(self):
        if isdefined(self.inputs.out_file):
            out_name = self.inputs.out_file
        else:
            base, ext = split_extension(
                os.path.basename(self.inputs.in_file))
            out_name = os.path.join(
                os.getcwd(), "{}_resize{}".format(base, ext))
        return out_name


class MRRegisterInputSpec(MRTrix3BaseInputSpec):
    in_file = File(exists=True, argstr='%s', mandatory=True, position=-3,
                   desc='input files')
    reference = File(exists=True, argstr='%s', mandatory=True, position=-2,
                     desc="The reference image to register to")
    out_file = File(genfile=True, argstr='%s', desc=(""), position=-1)


class MRRegisterOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc=(""))


class MRRegister(MRTrix3Base):
    """MRRegister"""

    _cmd = "mrregister"
    input_spec = MRRegisterInputSpec
    output_spec = MRRegisterOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = self._gen_outfilename()
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            gen_name = self._gen_outfilename()
        else:
            assert False
        return gen_name

    def _gen_outfilename(self):
        if isdefined(self.inputs.out_file):
            out_name = self.inputs.out_file
        else:
            base, ext = split_extension(
                os.path.basename(self.inputs.in_file))
            out_name = os.path.join(
                os.getcwd(), "{}_register{}".format(base, ext))
        return out_name


class MRThresholdInputSpec(MRTrix3BaseInputSpec):
    in_file = File(exists=True, argstr='%s', mandatory=True, position=-3,
                   desc='input files')
    out_file = File(genfile=True, argstr='%s', desc=(""), position=-1)


class MRThresholdOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc=(""))


class MRThreshold(MRTrix3Base):
    """MRThreshold"""

    _cmd = "mrthreshold"
    input_spec = MRThresholdInputSpec
    output_spec = MRThresholdOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = self._gen_outfilename()
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            gen_name = self._gen_outfilename()
        else:
            assert False
        return gen_name

    def _gen_outfilename(self):
        if isdefined(self.inputs.out_file):
            out_name = self.inputs.out_file
        else:
            base, ext = split_extension(
                os.path.basename(self.inputs.in_file))
            out_name = os.path.join(
                os.getcwd(), "{}_threshold{}".format(base, ext))
        return out_name


class FixelReorientInputSpec(MRTrix3BaseInputSpec):
    in_file = File(exists=True, argstr='%s', mandatory=True, position=-3,
                   desc='input files')
    out_file = File(genfile=True, argstr='%s', desc=(""), position=-1)


class FixelReorientOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc=(""))


class FixelReorient(MRTrix3Base):
    """FixelReorient"""

    _cmd = "fixelreorient"
    input_spec = FixelReorientInputSpec
    output_spec = FixelReorientOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = self._gen_outfilename()
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            gen_name = self._gen_outfilename()
        else:
            assert False
        return gen_name

    def _gen_outfilename(self):
        if isdefined(self.inputs.out_file):
            out_name = self.inputs.out_file
        else:
            base, ext = split_extension(
                os.path.basename(self.inputs.in_file))
            out_name = os.path.join(
                os.getcwd(), "{}_reorient{}".format(base, ext))
        return out_name



import os
import os.path as op
from copy import copy
import banana
from banana.requirement import fsl_req
from arcana.exceptions import ArcanaError
from arcana import Fileset, FilesetCollection


class BaseReference():

    frequency = 'per_study'

    def __init__(self, format, name=None):  # @ReservedAssignment
        self._name = name
        self._format = format
        self._study = None

    def bind(self, study):
        bound = copy(self)
        bound._study = study
        return bound

    @property
    def study(self):
        if self._study is None:
            raise ArcanaError(
                "Can't access study property as {} has not been bound"
                .format(self))
        return self._study

    @property
    def name(self):
        if self._name is None:
            raise ArcanaError(
                "Name for atlas hasn't been set, it should be set in when "
                "it is passed as a default")
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    def __hash__(self):
        return hash(self.name)

    @property
    def collection(self):
        return FilesetCollection(
            self.name,
            [Fileset.from_path(self.path, frequency=self.frequency)],
            format=self._format,
            frequency=self.frequency)

    @property
    def format(self):
        return self.collection.format

    @property
    def path(self):  # @UnusedVariable
        return NotImplementedError

    def __repr__(self):
        return '{}(name={})'.format(type(self).__name__, self._name)

    def __eq__(self, other):
        return self._name == other._name


class FslReferenceData(BaseReference):
    """
    Class to retrieve the path to an atlas shipped with a FSL installation

    Parameters
    ----------
    atlas_name : str
        Name of atlas or family of atlases
    resolution : str | float
        The resolution of the atlas to use. Can either be a fixed floating
        point value or the name of a parameter in the study to draw the
        value from
    dataset : str | None
        Name of the dataset (i.e. 'brain', 'brain_mask', 'eye_mask', 'edges')
        will be append to the atlas name using '_' as a delimeter
    sub_path : str
        Relative path to atlas directory from FSL 'data' directory
    """

    def __init__(self, atlas_name, format, name=None, resolution=1,  # @ReservedAssignment @IgnorePep8
                 dataset=None, sub_path=['standard']):
        super().__init__(format, name)
        self._atlas_name = atlas_name
        self._resolution = resolution
        self._dataset = dataset
        self._sub_path = tuple(sub_path.split('/') if isinstance(sub_path, str)
                               else sub_path)

    @property
    def path(self):
        # If resolution is a string then it is assumed to be a parameter name
        # of the study
        if isinstance(self._resolution, str):
            resolution = getattr(self.study, self._resolution)
        else:
            resolution = self._resolution
        full_atlas_name = '{}_{}mm'.format(self._atlas_name, resolution)
        if self._dataset is not None:
            full_atlas_name += '_' + self._dataset
        fsl_ver = self.study.environment.satisfy(fsl_req.v('5.0.8'))[0]
        if hasattr(self.study.environment, 'load'):
            self.study.environment.load(fsl_ver)
            fsl_dir = os.environ['FSLDIR']
            self.study.environment.unload(fsl_ver)
        else:
            fsl_dir = os.environ['FSLDIR']  # Static environments
        return op.join(fsl_dir, 'data', *self._sub_path,
                       full_atlas_name + '.nii.gz')

    def translate(self, substudy_spec):
        """
        Translate resolution parameter name if used to namespace of multi-study

        Parameters
        ----------
        substudy_spec : SubStudySpec
            The sub-study that the spec belongs to
        """
        if isinstance(self._resolution, str):
            self._resolution = substudy_spec.map(self._resolution)

    def __eq__(self, other):
        return (
            super().__eq__(other) and
            self._atlas_name == other._atlas_name and
            self._resolution == other._resolution and
            self._dataset == other._dataset and
            self._sub_path == other._sub_path)

    def __hash__(self):
        return (super().__hash__() ^
                hash(self._atlas_name) ^
                hash(self._resolution) ^
                hash(self._dataset) ^
                hash(self._sub_path))

    @property
    def _error_msg_loc(self):
        return "'{}' FSL atlas passed to '{}' in {} ".format(
            self._atlas_name, self.name, self.study)


class LocalReferenceData(BaseReference):
    """
    Several atlases used in the composite-vein analysis in the T2* study,
    stored within the banana package.

    Parameters
    ----------
    atlas_name : str
        Base name of the atlas file (i.e. without extension) in the 'atlases'
        directory
    """

    BASE_PATH = op.abspath(op.join(op.dirname(__file__), 'data'))

    def __init__(self, atlas_name, format, name=None):  # @ReservedAssignment
        super().__init__(format, name)
        self._atlas_name = atlas_name

    @property
    def path(self):  # @UnusedVariable
        return op.join(self.BASE_PATH, self._atlas_name + '.nii.gz')

    def __eq__(self, other):
        return (super().__eq__(other) and
                self._atlas_name == other._atlas_name)

    def __hash__(self):
        return (super().__hash__() ^
                hash(self._atlas_name))



from .base import MriStudy
from .dwi import DwiStudy
from .bold import BoldStudy
from .t1 import T1Study
from .t2 import T2Study
from .t2star import T2starStudy
# from .bold import BoldStudy
# from .ute import Utetudy
from .epi import EpiSeriesStudy



import os.path
import pydicom
import glob
import shutil
import errno
import subprocess as sp
from banana.interfaces.custom.dicom import DicomHeaderInfoExtraction
import numpy as np
import re
import datetime as dt


PHASE_IMAGE_TYPE = ['ORIGINAL', 'PRIMARY', 'P', 'ND']


# def xnat_motion_detection(xnat_id):
#
#     avail_scans = xnat_ls(xnat_id, datatype='scan')
#     print(avail_scans)
#
#     return avail_scans


def local_motion_detection(input_dir, pet_dir=None, pet_recon=None,
                           struct2align=None):

    scan_description = []
    dcm_files = sorted(glob.glob(input_dir+'/*.dcm'))
    if not dcm_files:
        dcm_files = sorted(glob.glob(input_dir+'/*.IMA'))
    else:
        dcm = True
    if not dcm_files:
        scan_description = [f for f in os.listdir(input_dir) if (not
                            f.startswith('.') and os.path.isdir(input_dir+f)
                            and 'motion_correction_results' not in f)]
        dcm = False
    else:
        dcm = True
    if not dcm_files and not scan_description:
        raise Exception('No DICOM files or folders found in {}'
                        .format(input_dir))
    try:
        os.mkdir(input_dir+'/work_dir')
        os.mkdir(input_dir+'/work_dir/work_sub_dir')
        os.mkdir(input_dir+'/work_dir/work_sub_dir/work_session_dir')
        working_dir = input_dir+'/work_dir/work_sub_dir/work_session_dir/'
        copy = True
    except OSError as e:
        if e.errno == errno.EEXIST:
            print ('Detected existing working directory. Assuming that a '
                   'previous process failed. Trying to restart it.')
            working_dir = input_dir+'/work_dir/work_sub_dir/work_session_dir/'
            copy = False
    if dcm:
        hdr = pydicom.read_file(dcm_files[0])
        name_scan = (
            str(hdr.SeriesNumber).zfill(2)+'_'+hdr.SeriesDescription)
        name_scan = name_scan.replace(" ", "_")
        scan_description = [name_scan]
        files = []
        for i, im in enumerate(dcm_files):
            hdr = pydicom.read_file(im)
            name_scan = (
                str(hdr.SeriesNumber).zfill(2)+'_'+hdr.SeriesDescription)
            name_scan = name_scan.replace(" ", "_")
            if name_scan in scan_description[-1]:
                files.append(im)
            elif name_scan not in scan_description[-1] and copy:
                if (os.path.isdir(
                        working_dir+scan_description[-1]) is False):
                    os.mkdir(working_dir+scan_description[-1])
                    for f in files:
                        shutil.copy(f, working_dir+scan_description[-1])
                files = [im]
                scan_description.append(name_scan)
            elif name_scan not in scan_description[-1] and not copy:
                files = [im]
                scan_description.append(name_scan)
            if i == len(dcm_files)-1 and copy:
                if (os.path.isdir(working_dir+scan_description[-1]) is
                        False):
                    os.mkdir(working_dir+scan_description[-1])
                    for f in files:
                        shutil.copy(f, working_dir+scan_description[-1])
    elif not dcm and copy:
        for s in scan_description:
            shutil.copytree(input_dir+s, working_dir+'/'+s)
        if pet_dir is not None:
            shutil.copytree(pet_dir, working_dir+'/pet_data_dir')
        if pet_recon is not None:
            shutil.copytree(pet_recon, working_dir+'/pet_data_reconstructed')
        if struct2align is not None:
            shutil.copy2(struct2align, working_dir+'/')

    phase_image_type, no_dicom = check_image_type(input_dir, scan_description)
    if no_dicom:
        print(('No DICOM files could be found in the following folders '
               'For this reason they will be removed from the analysis.\n{}'
               .format('\n'.join(x for x in no_dicom))))
        scan_description = [x for x in scan_description
                            if x not in no_dicom]
    if phase_image_type:
        print(('The following scans were found to be phase image '
               'For this reason they will be removed from the analysis.\n{}'
               .format('\n'.join(x for x in phase_image_type))))
        scan_description = [x for x in scan_description
                            if x not in phase_image_type]
    same_start_time = check_image_start_time(input_dir, scan_description)
    if same_start_time:
        print(('The following scans were found to have the same start time '
               'as other scans provided. For this reason they will be removed'
               ' from the analysis.\n{}'
               .format('\n'.join(x for x in same_start_time))))
        scan_description = [x for x in scan_description
                            if x not in same_start_time]

    return scan_description


def inputs_generation(scan_description, input_dir, siemens=False):

    ref = None
    umaps = None
    scan_description = sorted(scan_description)

    for i, scan in enumerate(scan_description, start=1):
        if i == 1:
            print('Available scans: ')
        print('{0} {1}'.format(i, scan))
    correct_ref = False
    while not correct_ref:
        ref = input("Please select the reference scan: ").split()
        if not ref:
            print ('A reference image must be provided!')
        elif ref and len(ref) > 1:
            print(('Only one reference can be provided, you selected {}.'
                   .format(len(ref))))
        else:
            correct_ref = True

    ref = scan_description[int(ref[0])-1]
    correct_ref_type = False
    while not correct_ref_type:
        ref_type = input(
            "Please enter the reference type ('t1' or 't2'): ").split()[0]
        if ref_type != 't1' and ref_type != 't2':
            print(('{} is not a recognized ref_type!The available '
                   'ref_types are t1 or t2.'.format(ref_type)))
        else:
            correct_ref_type = True
    t1s = input("Please select the T1 weighted scans: ").split()
    epis = input("Please select the EPI scans: ").split()
    dwi = input("Please select the DWI scans (including both b0 and"
                " main diffusion images): ").split()
    dwi = [scan_description[int(i)-1] for i in dwi]
    if dwi:
        dwis, unused_b0 = dwi_type_assignment(input_dir, dwi)
        if len(dwis) < len(dwi):
            raise Exception('The following DWI scan cannot be recognized as b0'
                            'or main diffusion scan. Please remove it from the'
                            'input directory or assign it to others:\n{}'
                            .format(' '.join(x for x in dwi if x not in
                                             dwis)))
        else:
            print ('The DWI images provided were assigned to the following '
                   'types: \n')
            for dwi in dwis:
                if dwi[-1] == '0':
                    print(('main diffusion image with multiple directions: '
                           '{}'.format(dwi[0])))
                elif dwi[-1] == '1':
                    print(('b0 image with same phase encoding direction '
                           'respect to the main dwi: {}'.format(dwi[0])))
                elif dwi[-1] == '-1':
                    print(('b0 image with opposite phase encoding direction '
                           'respect to the main dwi: {}'.format(dwi[0])))
            print('\n')
    if not siemens:
        utes = input("Please select the UTE scans: ").split()
        utes = [scan_description[int(i)-1] for i in utes]
        if utes:
            umaps = input("Please select the umap: ").split()
            umaps = [scan_description[int(i)-1] for i in umaps]

    t2s = input("Please select the all the other scans the do not belog to"
                " any of the previous classes: ").split()
    t1s = [scan_description[int(i)-1] for i in t1s]
    t2s = [scan_description[int(i)-1] for i in t2s]
    epis = [scan_description[int(i)-1] for i in epis]
    if unused_b0:
        print(('The following b0 images with different phase encoding '
               'direction respect to the main diffusion has been found. They '
               'will be treated as T2w:\n{}'
               .format('\n'.join(x for x in unused_b0))))
        t2s = t2s+unused_b0

    if siemens:
        return ref, ref_type, t1s, epis, t2s, dwis
    else:
        return ref, ref_type, t1s, epis, t2s, dwis, utes, umaps


def guess_scan_type(scans, input_dir):

    ref = None
    ref_type = None
    t1s = []
    t2s = []
    epis = []
    dwi_scans = []
    res_t1 = []
    res_t2 = []

    for scan in scans:
        sequence_name = None
        dcm_files = sorted(glob.glob(input_dir+'/'+scan+'/*.dcm'))
        if not dcm_files:
            dcm_files = sorted(glob.glob(input_dir+'/'+scan+'/*.IMA'))
        if not dcm_files:
            continue
        dicom = dcm_files[0]
        hd = pydicom.read_file(dicom)
        with open(dcm_files[0], 'rb') as f:
            for line in f:
                try:
                    line = line[:-1].decode('utf-8')
                except UnicodeDecodeError:
                    continue
                if 'tSequenceFileName' in line:
                    sequence_name = line.strip().split('\\')[-1].split('"')[0]
                    break

        if sequence_name is not None:
            if (('tfl' in sequence_name or
                    re.match('.*(ute|UTE).*', sequence_name)) or
                    (re.match('.*(t1|T1).*', scan) or
                     re.match('.*(ute|UTE).*', scan))):
                t1s.append(scan)
                res_t1.append([scan, float(hd.PixelSpacing[0])])
            elif 'bold' in sequence_name or 'asl' in sequence_name:
                epis.append(scan)
            elif 'diff' in sequence_name:
                dwi_scans.append(scan)
            else:
                t2s.append(scan)
                if 'gre' not in sequence_name:
                    res_t2.append([scan, float(hd.PixelSpacing[0])])
    dwis, unused_b0 = dwi_type_assignment(input_dir, dwi_scans)
    if unused_b0:
        print(('The following b0 images have different phase encoding '
               'direction respect to the main diffusion and/or the ped '
               'information could not be found in the image header. They '
               'will be treated as T2w:\n{}'
               .format('\n'.join(x for x in unused_b0))))
        t2s = t2s+unused_b0
    if res_t2:
        ref = [x[0] for x in res_t2 if x[1] <= 1]
        if ref:
            if len(ref) > 1:
                min_res = np.min([x[1] for x in res_t2])
                ref = [x[0] for x in res_t2 if x[1] == min_res]
            ref_type = 't2'
            ref = ref[0]
            t2s.remove(ref)
    if res_t1 and not ref:
        ref = [x[0] for x in res_t1 if x[1] <= 1]
        if ref:
            ref_type = 't1'
            ref = ref[0]
            t1s.remove(ref)
    if ref:
        print(('\nChosen reference image: {0} \nReference type: {1}\n'
               .format(ref, ref_type)))
        if t1s:
            print ('The following scans were identified as T1 weighted: \n')
            for t1 in t1s:
                print(('{}'.format(t1)))
            print('\n')
        if epis:
            print ('The following scans were identified as 4D BOLD or ASL '
                   'images:\n')
            for epi in epis:
                print(('{}'.format(epi)))
            print('\n')
        if dwis:
            print ('The following scans were identified as DWI: \n')
            for dwi in dwis:
                if dwi[-1] == '0':
                    print(('main diffusion image with multiple directions: '
                           '{}'.format(dwi[0])))
                elif dwi[-1] == '1':
                    print(('b0 image with same phase encoding direction '
                           'respect to the main dwi: {}'.format(dwi[0])))
                elif dwi[-1] == '-1':
                    print(('b0 image with opposite phase encoding direction '
                           'respect to the main dwi: {}'.format(dwi[0])))
            print('\n')
        if t2s:
            print ('The following scans were identified as not belonging to '
                   'any of the previous classes: \n')
            for t2 in t2s:
                print(('{}'.format(t2)))
            print('\n')

        assigned = [ref]+t1s+t2s+[x[0] for x in dwis]+epis
        not_assigned = [x for x in scans if x not in assigned]
        if not_assigned:
            print (
                'The following scans could not be assigned to any class.'
                ' If DWI images are present, this might be due to the '
                'inhability of the pipeline to find phase encoding information'
                'in the header. \n')
            for na in not_assigned:
                print(('{}'.format(na)))
            print('\n')

        check_guess = input(
            "Please type 'yes' if the grouping is correct, otherwise 'no'. "
            "If it is not correct then you will be prompted to manually group "
            "all the scans into the different classes: ").split()
        if check_guess[0] == 'yes':
            inputs = [ref, ref_type, t1s, epis, t2s, dwis]
        else:
            inputs = []
    else:
        print ('Reference image could not be identified from header '
               'information. You will be prompted to manually group all '
               'the scans into different classes.')
        inputs = []

    return inputs


def dwi_type_assignment(input_dir, dwi_images):

    main_dwi = []
    b0 = []
    dwis = []
    unused_b0 = []

    for dwi in dwi_images:
        cmd = 'mrinfo {0}'.format(input_dir+'/'+dwi)
        info = (sp.check_output(cmd, shell=True)).decode('utf-8')
        info = info.strip().split('\n')
        for line in info:
            if 'Dimensions:' in line:
                dim = line.split('Dimensions:')[-1].strip().split('x')
                break
        hd_extraction = DicomHeaderInfoExtraction()
        hd_extraction.inputs.dicom_folder = input_dir+'/'+dwi
        dcm_info = hd_extraction.run()

        if dcm_info.outputs.pe_angle and dcm_info.outputs.ped:
            if len(dim) == 4:
                main_dwi.append(
                    [dwi, np.trunc(float(dcm_info.outputs.pe_angle)),
                     dcm_info.outputs.ped])
            else:
                b0.append([
                    dwi, np.trunc(float(dcm_info.outputs.pe_angle)),
                    dcm_info.outputs.ped])
        else:
            print ('Could not find phase encoding information from the'
                   'dwi images header. Distortion correction will not '
                   'be performed.')
            if len(dim) == 4:
                main_dwi.append(
                    [dwi, '', ''])
            else:
                b0.append([
                    dwi, '-2', '-2'])

    for i in range(len(main_dwi)):
        if main_dwi[i][2]:
            dwis.append([main_dwi[i][0], '0'])
        ped_main = main_dwi[i][2]
        for j in range(len(b0)):
            ped_b0 = b0[j][2]
            if ped_b0 == ped_main:
                if main_dwi[i][1] == b0[j][1] and (j == i or j == i+1):
                    dwis.append([b0[j][0], '1'])
                elif main_dwi[i][1] != b0[j][1] and (j == i or j == i+1):
                    dwis.append([b0[j][0], '-1'])
            else:
                unused_b0.append(b0[j][0])
#         if not b0_found:
#             raise Exception(
#                 'The phase encoding direction between the main'
#                 'DWI and the provided b0 images is not the '
#                 'same! Please check.')

    return dwis, unused_b0


def check_image_type(input_dir, scans):

    toremove = []
    nodicom = []
    for scan in scans:
        dcm_file = None
        try:
            dcm_file = sorted(glob.glob(input_dir+'/'+scan+'/*.dcm'))[0]
        except IndexError:
            try:
                dcm_file = sorted(glob.glob(
                    input_dir+'/'+scan+'/*.IMA'))[0]
            except IndexError:
                nodicom.append(scan)
        if dcm_file is not None:
            try:
                hd = pydicom.read_file(dcm_file)
                im_type = hd['0008', '0008'].value
                if im_type == PHASE_IMAGE_TYPE:
                    toremove.append(scan)
            except:
                print(('{} does not have the image type in the header. It will'
                       ' be removed from the analysis'.format(scan)))

    return toremove, nodicom


def check_image_start_time(input_dir, scans):

    start_times = []
    toremove = []
    for scan in scans:
        try:
            scan_number = scan.split('-')[0].zfill(3)
            hd_extraction = DicomHeaderInfoExtraction()
            hd_extraction.inputs.dicom_folder = input_dir+'/'+scan
            dcm_info = hd_extraction.run()
            start_times.append([dcm_info.outputs.start_time, scan_number,
                                scan])
        except:
            print(('This folder {} seems to not contain DICOM files. It will '
                   'be ingnored.'.format(scan)))
    start_times = sorted(start_times)
    for i in range(1, len(start_times)):
        diff = ((dt.datetime.strptime(start_times[i][0], '%H%M%S.%f') -
                dt.datetime.strptime(start_times[i-1][0], '%H%M%S.%f'))
                .total_seconds())
        if diff < 5:
            toremove.append(start_times[i][-1])

    return toremove


def md_cleanup(input_dir, work_dir, project_id, sub_id, session_id):

    shutil.move(os.path.join(
        input_dir, project_id, sub_id, session_id,
        'motion_detection_motion_detection_output'), input_dir)
    os.rename(
        os.path.join(input_dir, 'motion_detection_motion_detection_output'),
        os.path.join(input_dir, 'motion_detection_output'))
    shutil.rmtree(work_dir)
    shutil.rmtree(os.path.join(input_dir, project_id))



from .base import PetStudy
from arcana.data import FilesetSpec
from arcana.study.base import StudyMetaClass
from banana.interfaces.custom.pet import SUVRCalculation
from banana.file_format import (nifti_gz_format)
import os

template_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__).split('arcana')[0],
                 'arcana', 'reference'))


class StaticPetStudy(PetStudy, metaclass=StudyMetaClass):

    add_data_specs = [
        FilesetSpec('pet_image', nifti_gz_format),
        FilesetSpec('base_mask', nifti_gz_format),
        FilesetSpec('SUVR_image', nifti_gz_format, 'suvr_pipeline')]

    def suvr_pipeline(self, **kwargs):

        pipeline = self.new_pipeline(
            name='SUVR',
            desc=('Calculate SUVR image'),
            citations=[],
            **kwargs)

        pipeline.add(
            'SUVR',
            SUVRCalculation(),
            inputs={
                'volume': ('registered_volume', nifti_gz_format),
                'base_mask': ('base_mask', nifti_gz_format)},
            outputs={
                'SUVR_image': ('SUVR_file', nifti_gz_format)})

        return pipeline

    def _ica_inputs(self):
        pass
